"""
In the previous sections you we had seen all type of ussd screen and as you had
seen all screens are saved in one model but the logistic of creating different screens
is defined by the apis.

In this section we are going to see the apis that create the different types of
screens.

"""


class InputScreen(object):
    """
    This screen contains only text and input.

    Once the user enters input, the user is directed to the next screen.
    Two things to note. we need to know next screen to navigate to and
    we also need to store the users input.

    Its created be hitting this view with the following fields:

    * text
        *type*: string

        *description*:  representing whats going to be displayed to the user.

    * next_ussd_screen
        *type*: int -> this should be an int representing the pk of an existing ussd screen

        *description*:  represents the next screen to navigate to if user enters input

    * session_key
        *type*: string

        *description*: uses input will need to be saved in the session,
        it will use this key to save the input

    * initial_screen
        *type*: boolean

        * description*: if its true, it becomes the initial screen and
        the parent of all screens, in that ussd workflow
    """
    pass


class MenuScreen(object):
    """
    This screen contains text and list of options to select from
    and input field.

    The user is supposed to select one of the options provided.
    each option should have next ussd screen to direct to.

    This is screen is created by hitting this view with the following
    fields:

    * text
        *type*: string

        *description*:  representing whats going to be displayed to the user.

    * menu_options
        *type*: dict
            contain the following content

            * menu_text
                *type*: string

                *description*: this is text going to be displayed when showing options

            * menu_input
                *type*: string or int

                *description*: this is value the user is going to select when choosing
                a option

            * menu_index
                *type*: int -> its optional if menu_input is a int

                *description*: Menu options displayed to users are
                arranged depending on this field

                *constrain*: menu_index should be unique per each screen.
                you can't have two menu options with the same menu_index

            * next_ussd_screen
                *type*: int -> this should be an int representing the pk of an
                 existing ussd screen

                *description*:  represents the next screen to navigate to if
                user enters input
    """


class ListScreen(object):
    """
    This screen contains text and list of options to select from
    and input field.

    The user is supposed to select one of the options provided.
    each option should have next ussd screen to direct to.

    The difference between menu screen and list screen is the options displayed
    in in ListScreen can be dynamic options (i.e the are generated from somewhere)

    This screen is created by hitting this view with the following fields:

    * text
        *type*: string

        *description*:  representing whats going to be displayed to the user.

    * list
        comming soon

    * session_key
        *type*: string

        *description*: uses input will need to be saved in the session,
        it will use this key.

        NB. user input will only be saved in the session with this key only if
        the use selected the generated options not menu_options

    * next_ussd_screen
        *type*: int -> this should be an int representing the pk of an
        existing ussd screen

        *description*:  represents the next screen to navigate to if user enters input

        NB. user will only be directed to this screen only if the user selected the
        generated options not menu_options

    * menu_options
        *type*: MenuOption as described in menu screen

    """


class QuitScreen(object):
    """
    Its the easiest screen to create.
    It displays text only to the user.

    Its the screen responsible for ending session.

    To create this screen you hit this view with the following fields:

    * text
        *type*: string

        *description*:  representing whats going to be displayed to the user.

    """
    pass


class RouterScreen(object):
    """
    This screen is not displayed to the users.

    The purpose of this screen is to route uses to a screen depending on
    the route_option specified.

    This screen is created by hitting this view with the following fields:

    *route_options
        comming soon
    """


class HttpRequest(object):
    """
    This screen is not displayed to the users.

    The purpose of this screen is to make a http request and save
    the response in the session

    This screen is created by hitting this view with the following fields:

    * session_key
        - This key is used to save the response in the session

    * others comming soon
    """
    pass



from ussd.core import UssdHandlerAbstract, load_yaml
from rest_framework import serializers
from ussd.screens.serializers import NextUssdScreenSerializer
import staticconf


class VariableDefinition(serializers.Serializer):
    file = serializers.CharField()
    namespace = serializers.CharField(max_length=100)


class ValidateResposeSerialzier(serializers.Serializer):
    expression = serializers.CharField(max_length=255)


class UssdReportSessionSerializer(serializers.Serializer):
    session_key = serializers.CharField(max_length=100)
    validate_response = serializers.ListField(
        child=ValidateResposeSerialzier()
    )
    request_conf = serializers.DictField()



class InitialScreenSerializer(NextUssdScreenSerializer):
    variables = VariableDefinition(required=False)
    create_ussd_variables = serializers.DictField(default={})
    default_language = serializers.CharField(required=False,
                                             default="en")
    ussd_report_session = UssdReportSessionSerializer(required=False)


class InitialScreen(UssdHandlerAbstract):
    """This screen is mandatory in any customer journey.
    It is the screen all new ussd session go to.

    example of one

        .. code-block:: yaml

            initial_screen: enter_height

            first_screen:
            type: quit
            text: This is the first screen

    Its is also used to define variable file if you have one.
    Example when defining variable file

        .. code-block:: yaml

            initial_screen:
                screen: screen_one
                variables:
                    file: /path/of/your/variable/file.yml
                    namespace: used_to_save_the_variable
    Sometimes you want to send ussd session to some 3rd party application when
    the session has been terminated.
    
    We can easily do that at end of session i.e quit screen, But for those
    scenarios where session is terminated by user or mno we don't know that 
    unless the mno send us a request. 
    
    Most mnos don't send notifier 3rd party application about the session being
    dropped. The work around we use is schedule celery task to run after 
    15 minutes ( by that time we know there is no active session)
    
    Below is an example of how to schedule a ussd report session after 15min
        
    
    example:
        
        .. code-block:: yaml
        
            initial_screen:
                type: initial_screen
                next_screen: screen_one
                ussd_report_session:
                    session_key: reported
                    retry_mechanism:
                        max_retries: 3
                    validate_response:
                        - expression: "{{reported.status_code}} == 200"
                    request_conf:
                        url: localhost:8006/api
                        method: post
                        data:
                            ussd_interaction: "{{ussd_interaction}}"
                            session_id: "{{session_id}}"
                    async_parameters:
                        queue: report_session
                        countdown: 900
                    
        
        Lets explain the variables in ussd_report_session
            - session_key ( Mandatory )
                response of ussd report session would be saved under that key
                in session store
                
            - request_conf ( Mandatory )
                Those are the parameters to be used to make request to 
                report ussd session
            
            - validate_response ( Mandatory )
                After making ussd report request the framework will evaluate 
                your options and if one of them is valid it would 
                mark session as posted (This is used to avoid double ussd 
                submission)
                
            - retry_mechanism ( Optional )
                After validating your response and all of them fail 
                we will go ahead and retry if this field is active. 
                
            - async_parameters ( Optional )
                This is are the parameters used to make ussd request
            
            
                
    """
    screen_type = "initial_screen"

    serializer = InitialScreenSerializer

    def handle(self):

        if isinstance(self.screen_content, dict):
            if self.screen_content.get('variables'):
                self.load_variable_files()

            # create ussd variables defined int the yaml
            self.create_variables()

            # set default language
            self.set_language()

            next_screen = self.screen_content['next_screen']

            # call report session
            if self.screen_content.get('ussd_report_session'):
                self.fire_ussd_report_session_task(self.initial_screen,
                                                   self.ussd_request.session_id
                                                   )
        else:
            next_screen = self.screen_content
        return self.ussd_request.forward(next_screen)

    def create_variables(self):
        for key, value in \
                self.screen_content.get('create_ussd_variables', {}). \
                        items():
            self.ussd_request.session[key] = \
                self.evaluate_jija_expression(value,
                                              lazy_evaluating=True,
                                              session=self.ussd_request.session
                                              )

    def load_variable_files(self):
        variable_conf = self.screen_content['variables']
        file_path = variable_conf['file']
        namespace = variable_conf['namespace']

        # check if it has been loaded
        if not namespace in \
                staticconf.config.configuration_namespaces:
            load_yaml(file_path, namespace)

        self.ussd_request.session.update(
            staticconf.config.configuration_namespaces[namespace].
            configuration_values
        )

    def set_language(self):
        self.ussd_request.session['default_language'] = \
            self.screen_content.get('default_language', 'en')



from ussd.core import register_filter
from datetime import datetime
import calendar


@register_filter
def day(date_time):
    return date_time.day


@register_filter
def month(date_time):
    return date_time.month


@register_filter
def year(date_time):
    return date_time.year


@register_filter
def month_name(no_month):
    if isinstance(no_month, datetime):
        no_month = no_month.month
    return calendar.month_name[no_month]


@register_filter
def day_name(date_time):
    return date_time.strftime("%A")


@register_filter
def add_month(month, months_to_add):
    next_month = month + months_to_add

    if next_month > 12:
        next_month -= 12
    elif next_month < 1:
        next_month += 12
    return next_month


@register_filter
def strip(strftime, strptime):
    return datetime.strptime(
        strftime,
        strptime
    )



from ussd.core import UssdHandlerAbstract
from ussd.screens.serializers import UssdBaseSerializer, \
    NextUssdScreenSerializer
from rest_framework import serializers


class UpdateSessionExpressionSerializer(serializers.Serializer):
    expression = serializers.CharField(max_length=255, required=False)
    key = serializers.CharField(max_length=255)
    value = serializers.CharField(max_length=255)


class UpdateSessionSerializer(UssdBaseSerializer, NextUssdScreenSerializer):
    values_to_update = serializers.ListField(
        child=UpdateSessionExpressionSerializer()
    )


class UpdateSessionScreen(UssdHandlerAbstract):
    """
    This screen is invisible to the user. Sometimes you may want to save
    something to the session to use later in other screens.

    Fields used to create this screen:
        1. next_screen
            The screen to go after the session has been saved

        2. values_to_update
            This section defines the session to be saved.

            Inside this section should define the following fields

            i. key
                    the key to be used to save
            ii. value
                    the value to store with the key above
            iii. expression
                    sometimes you want a condition before you can save data in
                    section

    Example:
        .. literalinclude:: .././ussd/tests/sample_screen_definition/valid_update_session_screen_conf.yml

    """
    screen_type = "update_session_screen"
    serializer = UpdateSessionSerializer

    def handle(self):

        loop_items = self.get_loop_items()

        values_to_update = self.screen_content["values_to_update"]

        for item in loop_items:
            # update extra context
            extra_context = {
                "item": item
            }
            if isinstance(loop_items, dict):
                extra_context.update(
                    dict(
                        key=item,
                        value=loop_items[item],
                        item={item: loop_items[item]}
                    )
                )

            for update_value in values_to_update:
                if not (update_value.get('expression') and
                            self.evaluate_jija_expression(
                                update_value['expression'],
                                session=self.ussd_request.session,
                                extra_context=extra_context)):
                    continue

                # evaluate key
                key = update_value['key'] \
                    if not UssdHandlerAbstract._contains_vars(
                    update_value['key']) \
                    else self.evaluate_jija_expression(
                    update_value['key'],
                    session=self.ussd_request.session,
                    extra_context=extra_context
                )

                value = self.evaluate_jija_expression(
                    update_value['value'],
                    session=self.ussd_request.session,
                    extra_context=extra_context
                ) or update_value['value']

                # save them in the session store
                self.ussd_request.session[key] = value
        return self.ussd_request.forward(
            self.screen_content['next_screen']
        )




"""
This test initial screen
"""
from ussd.tests import UssdTestCase


class TestInitialHandler(UssdTestCase.BaseUssdTestCase):
    validation_error_message = {
        'initial_screen': {
            'next_screen': ['invalid_screen is missing in ussd journey'],
            'variables': {'file': ['This field is required.'],
                          'namespace': ['This field is required.']},
            'ussd_report_session': {
                "session_key": ['This field is required.'],
                'validate_response': ['This field is required.'],
                'request_conf': ['This field is required.'],
            }
        }
    }



from ussd.core import UssdHandlerAbstract
from ussd.screens.serializers import NextUssdScreenSerializer
from rest_framework import serializers
from ussd.tasks import http_task


class HttpScreenConfSerializer(serializers.Serializer):
    method = serializers.ChoiceField(
        ("post", "get", "put", "delete")
    )
    url = serializers.CharField(max_length=255)


class HttpScreenSerializer(NextUssdScreenSerializer):
    session_key = serializers.CharField()
    synchronous = serializers.BooleanField(required=False)
    http_request = HttpScreenConfSerializer()


class HttpScreen(UssdHandlerAbstract):
    """
    This screen is invisible to the user. Its used if you want to make an
    api call. Its very if you want to make a api call so that you can show
    the user the results in the next screen.

    For instance you can make call for balance check using this screen.
    And display the balance in the next screen.

    Fields used to create this screen:

        1. http_request
            This field contains all the fields used to make http request.
            It contains the following fields:
                a. method
                    This is the request method to use.
                        either: get, post, put, delete
                b. url
                    This is the url to be used to make the api call
                c. And all the parameters python request module would accept

                you will example below

        2. session_key
            In this screen the api call is expected to return json body.
            The json body is saved in session using this session_key

        3. synchronous (optional defaults to true)
           This defines the nature of the api call. If its asynchronous the
           request will be made later in celery task.

        4. next_screen
            After the api call has been made or been scheduled to celery task
            ussd request is forwarded to this next_screen

    Examples of router screens:

        .. literalinclude:: .././ussd/tests/sample_screen_definition/valid_http_screen_conf.yml
    """
    screen_type = "http_screen"
    serializer = HttpScreenSerializer

    def handle(self):
        http_request_conf = self.render_request_conf(
            self.ussd_request.session,
            self.screen_content['http_request']
        )

        if self.screen_content.get('synchronous', False):
            http_task.delay(request_conf=http_request_conf)
        else:
            self.make_request(
                http_request_conf=http_request_conf,
                response_session_key_save=self.screen_content['session_key'],
                session=self.ussd_request.session,
                logger=self.logger
            )
        return self.ussd_request.forward(self.screen_content['next_screen'])



from ussd.core import register_filter
from jinja2.runtime import Undefined


@register_filter
def append(list_a, obj):
    if isinstance(list_a, Undefined) or list_a is None or \
            (len(list_a) == 1 and isinstance(list_a[0], Undefined)):
        list_a = []
    list_a.append(obj)
    return list_a



from ussd.core import UssdHandlerAbstract
from ussd.screens.serializers import UssdBaseSerializer, \
    NextUssdScreenSerializer
from rest_framework import serializers
from ussd.tests import UssdTestCase
from ussd.core import UssdResponse


class InvalidCustomHandler(object):
    pass


class SampleCustomHandler1(UssdHandlerAbstract):
    abstract = True  # don't register custom classes
    @staticmethod
    def show_ussd_content():  # This method doesn't have to be static
        # Do anything custom here.
        return UssdResponse("This is a custom Handler1")

    def handle_ussd_input(self, ussd_input):
        # Do anything custom here
        print(ussd_input)  # pep 8 for the sake of using it.
        return self.ussd_request.forward('custom_screen_2')


class SampleSerializer(UssdBaseSerializer, NextUssdScreenSerializer):
        input_identifier = serializers.CharField(max_length=100)


class SampleCustomHandlerWithSerializer(UssdHandlerAbstract):
    abstract = True  # don't register custom classes
    serializer = SampleSerializer

    @staticmethod
    def show_ussd_content():  # This method doesn't have to be static
        return "Enter a digit and it will be doubled on your behalf"

    def handle_ussd_input(self, ussd_input):
        self.ussd_request.session[
            self.screen_content['input_identifier']
        ] = int(ussd_input) * 2

        return self.ussd_request.forward(
            self.screen_content['next_screen']
        )


class TestCustomHandler(UssdTestCase.BaseUssdTestCase):
    validation_error_message = dict(
        custom_screen_0=dict(
            screen_obj=['This field is required.']
        ),
        custom_screen_1=dict(
            screen_obj=["Screen object should be of type UssdHandlerAbstract"]
        ),
        custom_screen_2=dict(
            screen_obj=['Class does not exist']
        ),
        custom_screen_3=dict(
            next_screen=['This field is required.'],
            input_identifier=['This field is required.']
        )
    )

    def test(self):
        ussd_client = self.ussd_client()

        self.assertEqual(
            "This is a custom Handler1",
            ussd_client.send('')  # dial in
        )

        self.assertEqual(
            "Enter a digit and it will be doubled on your behalf",
            ussd_client.send('1')  # goes to the next screen
        )

        self.assertEqual(
            "Your custom screen has modified your input to 18",
            ussd_client.send('9')  # enter number to double
        )


from ussd.tests import UssdTestCase
from ussd.core import ussd_session


class TestScreensUsingVariable(UssdTestCase.BaseUssdTestCase):
    validate_ussd = False

    def get_ussd_client(self):
        return self.ussd_client(
            generate_customer_journey=False,
            extra_payload={
                "customer_journey_conf": "sample_using_variables.yml"
            }
        )

    def test_initial_variables_are_created(self):
        ussd_client = self.get_ussd_client()

        # dial in
        ussd_client.send('')

        session = ussd_session(ussd_client.session_id)

        # check ussd variables were created
        expected_variables = {"name": "mwas", "hobbies": [],
                              "environment": "sample_variable_two"}

        for key, value in expected_variables.items():
            self.assertTrue(key in session)
            self.assertEqual(session[key], value)

    def test_using_variable(self):
        client = self.get_ussd_client()

        # dial in
        response = client.send('1')

        self.assertEqual(
            "Displaying variable in the first config bar_variable_one. "
            "press one to show variable in config two\n1. Continue\n",
            response
        )

        # select one to continue
        response = client.send('1')

        self.assertEqual(
            "Displaying variable in the second config that has overridden "
            "the second config foo_variable_two.\n1. continue\n",
            response
        )

    def excute_following_inputs(self, inputs, expected_response):
        client = self.get_ussd_client()
        response = ''
        for i in inputs:
            response = client.send(i)
        self.assertEqual(expected_response, response)
        return client

    def test_using_variable_in_router_screen(self):

        # sequence of input to reach router screen
        inputs = ['', '1', '1']

        # screen that directs to router screen
        screen_three_expected_text = "Enter your any number we compare it " \
                                     "aganist a variable " \
                                     "set in the variables.\n"

        client = self.excute_following_inputs(inputs,
                                              screen_three_expected_text)

        # enter number equal to variable number
        response = client.send(50)

        self.assertEqual(
            "The number you have entered is equal to the variable 50",
            response
        )

        client = self.excute_following_inputs(inputs,
                                              screen_three_expected_text)

        # enter number greater than input
        self.assertEqual(
            "The number you have entered is greater than the variable 50",
            client.send('55')
        )

        client = self.excute_following_inputs(inputs,
                                              screen_three_expected_text)

        # enter number less than input
        self.assertEqual(
            "The number you have entered is less than "
            "the variable 50. "
            "Displaing a variable that has not been defined \n",
            client.send('40')
        )



from ussd.tests import UssdTestCase
from datetime import datetime
import calendar
from freezegun import freeze_time


@freeze_time(datetime.now())
class TestScreensUsingFilters(UssdTestCase.BaseUssdTestCase):

    validate_ussd = False

    def get_ussd_client(self):
        return self.ussd_client(
            generate_customer_journey=False,
            extra_payload={
                "customer_journey_conf": "sample_using_date_filters.yml"
            }
        )

    def test_using_filters(self):
        now = datetime.now()
        client = self.get_ussd_client()
        # dial in
        response = client.send('1')

        self.assertEqual(
            "The date today is {now}. We are on the {now_month} th month "
            "and we are on year {now_year}. Month in words is "
            "{now_month_name}. And day in words "
            "{now_day_name}. And next month {three_months}\n"
            "Testing striping date. After striping this "
            "date 2017-01-20 we should get the year, month and day. "
            "The day is {strip_day_name}\n".format(
                now=now,
                now_month=now.month,
                now_year=now.year,
                now_month_name=calendar.month_name[now.month],
                now_day_name=now.strftime("%A"),
                three_months=calendar.month_name[now.month + 3],
                strip_day_name=20
            ),
            response
        )



from ussd.tests import UssdTestCase
from unittest import mock
from django.http.response import JsonResponse, HttpResponse
from django.test.utils import override_settings


@override_settings(
        CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
        CELERY_ALWAYS_EAGER=True,
        BROKER_BACKEND='memory'
    )
class TestHttpScreen(UssdTestCase.BaseUssdTestCase):
    validation_error_message = dict(
        screen_name="Screen not available",
        http_invalid_screen=dict(
            next_screen=['This field is required.'],
            session_key=['This field is required.'],
            http_request=['This field is required.']
        ),
        http_screen_invalid_method=dict(
            http_request=dict(
                method=['"done" is not a valid choice.'],
            )
        ),
        http_screen_invalid_synchronous=dict(
            synchronous=['"not boolean" is not a valid boolean.']
        )
    )

    @mock.patch("ussd.core.requests.request")
    def test(self, mock_request):
        mock_response = JsonResponse({"balance": 250})
        mock_request.return_value = mock_response
        ussd_client = self.ussd_client()

        self.assertEqual(
            "Testing response is being saved in "
            "session status code is 200 and "
            "balance is 250 and full content {'balance': 250}.\n",
            ussd_client.send('')
        )
        expected_calls = [
            mock.call(
                method='get',
                url="http://localhost:8000/mock/balance",
                params=dict(
                    phone_number="200",
                    session_id=ussd_client.session_id
                ),
                verify=False,
                headers={"content-type": "application/json",
                         "user-agent": "my-app/0.0.1"}
            ),
            mock.call(
                method="get",
                url="http://localhost:8000/mock/balance/200/"
            ),
            mock.call(
                method='post',
                url='http://localhost:8000/mock/balance',
                params=dict(
                    phone_numbers=[200, 201, 202],
                    session_id=ussd_client.session_id
                ),
                verify=True,
                timeout=30,
                headers={"content-type": "application/json"}
            )
        ]
        # test requests that were made
        mock_request.assert_has_calls(expected_calls)

    @mock.patch("ussd.screens.http_screen.http_task")
    @mock.patch("ussd.tasks.requests.request")
    def test_async_workflow(self, mock_request, mock_http_task):
        mock_response = JsonResponse({"balance": 257})
        mock_request.return_value = mock_response

        ussd_client = self.ussd_client()
        ussd_client.send('')

        # check http_task is called
        mock_http_task.delay.assert_called_once_with(
            request_conf=dict(
                method='get',
                url="https://localhost:8000/mock/submission",
                params={'phone_number': '200',
                        'session_id': ussd_client.session_id}
            )
        )

    @mock.patch("ussd.core.requests.request")
    def test_json_decoding(self, mock_request):
        mock_response = HttpResponse("Balance is 257")
        mock_request.return_value = mock_response

        ussd_client = self.ussd_client()
        ussd_client.send('')

        self.assertEqual(
            "Testing response is being saved in "
            "session status code is 200 and "
            "balance is  and full content Balance is 257.\n",
            ussd_client.send('')
        )





from ussd.tests import UssdTestCase
from ussd.core import ussd_session


class TestInputHandler(UssdTestCase.BaseUssdTestCase):

    validation_error_message = dict(
            enter_height={
                "validators": {
                    'text': ['This field is required.']
                },
                "next_screen": ['This field is required.']
            },
            enter_age={
                "input_identifier": ['This field is required.'],
                "next_screen": ['thank_you_screen is missing in ussd journey'],
                "options": ['Expected a list of items but got type "bool".']
            },
            show_information={
                "type": ['Invalid screen type not supported']
            },
            hidden_fields={
                "initial_screen": ["This field is required."]
            }
        )

    def test_showing_screen_content(self):
        ussd_client = self.ussd_client()

        self.assertEqual(
            "Enter your height\n",
            ussd_client.send('')  # dial in
        )

        self.assertEqual(
            "Enter your age\n1. back\n",
            ussd_client.send('5')  # enter height
        )

        # confirm height entered is 5 before going back
        self.assertEqual(
            ussd_session(ussd_client.session_id)["height"],
            '5'
        )

        self.assertEqual(
            "Enter your height\n",
            ussd_client.send('1')  # enter one to go back
        )

        ussd_client.send('6')  # enter height

        self.assertEqual(
            "Your age is 24 and your height is 6.\n"
            "Enter anything to go back to the first screen\n",
            ussd_client.send('24') # enter age
        )

    def test_multilanguage_support(self):

        ussd_client = self.ussd_client(language='sw')


        # Dial in
        response = ussd_client.send('1')

        self.assertEqual(
            "Weka ukubwa lako\n",
            response
        )

        response = ussd_client.send('7')

        self.assertEqual(
            "Weka miaka yako\n1. rudi\n",
            response
        )

        response = ussd_client.send('23')

        self.assertEqual(
            "Miaka yako in 23 na ukubwa wako in 7.\n"
            "Weka kitu ingine yoyote unende "
            "kwenye screen ya kwanza\n",
            response
        )

    def test_input_validation(self):

        ussd_client = self.ussd_client()

        # dial in
        ussd_client.send('')

        # enter invalid height
        response = ussd_client.send('mwas')

        # should get a invalid error message
        self.assertEqual(
            "Enter number between 1 and 7\n",
            response
        )

        # enter valid height
        response = ussd_client.send('6')

        self.assertEqual(
            "Enter your age\n1. back\n",
            response
        )

        # enter invalid age greater thatn 100
        response = ussd_client.send('150')

        self.assertEqual(
            "Number over 100 is not allowed\n",
            response
        )

        # enter a valid age
        response = ussd_client.send('23')

        self.assertEqual(
            "Your age is 23 and your height is 6.\n"
            "Enter anything to go back to the first screen\n",
            response
        )



import uuid
import requests
import staticconf
from django.test import LiveServerTestCase, TestCase
from django.urls import reverse
from ussd.core import UssdView, load_yaml
from ussd.tests.sample_screen_definition import path


class UssdTestCase(object):
    """
    this contains two test that are required in each screen test case
    """

    class BaseUssdTestCase(LiveServerTestCase):
        validate_ussd = True

        def setUp(self):
            file_yml = self.__module__.split('.')[-1]. \
                           replace('test_', '') + '_conf.yml'
            self.valid_yml = 'valid_' + file_yml
            self.invalid_yml = 'invalid_' + file_yml
            self.namespace = self.__module__.split('.')[-1]
            self.maxDiff = None

            super(UssdTestCase.BaseUssdTestCase, self).setUp()

        def _test_ussd_validation(self, yaml_to_validate, expected_validation,
                                  expected_errors):

            if self.validate_ussd:
                namespace = self.namespace + str(expected_validation)
                file_path = path + '/' + yaml_to_validate
                load_yaml(file_path, namespace)
                ussd_screens = staticconf.config. \
                    get_namespace(namespace). \
                    get_config_values()

                is_valid, error_message = UssdView.validate_ussd_journey(
                    ussd_screens)

                self.assertEqual(is_valid, expected_validation, error_message)

                self.assertDictEqual(error_message,
                                     expected_errors)

        def testing_valid_customer_journey(self):
            self._test_ussd_validation(self.valid_yml, True, {})

        def testing_invalid_customer_journey(self):

            self._test_ussd_validation(self.invalid_yml, False,
                                       getattr(self,
                                               "validation_error_message",
                                               {}))

        def ussd_client(self, generate_customer_journey=True, **kwargs):
            class UssdTestClient(object):
                def __init__(self, host, session_id=None, phone_number=200,
                             language='en', extra_payload={},
                             service_code="test"):
                    self.phone_number = phone_number
                    self.language = language
                    self.session_id = session_id \
                        if session_id is not None \
                        else str(uuid.uuid4())
                    self.url = host + reverse('africastalking_url')
                    self.extra_payload = extra_payload
                    self.service_code = service_code

                def send_(self, ussd_input):
                    payload = {
                        "sessionId": self.session_id,
                        "text": ussd_input,
                        "phoneNumber": self.phone_number,
                        "serviceCode": self.service_code,
                        "language": self.language
                    }
                    payload.update(self.extra_payload)
                    response = requests.post(
                        url=self.url,
                        data=payload
                    )
                    return response

                def send(self, ussd_input):
                    return self.send_(ussd_input).content.decode()

            if generate_customer_journey:
                customer_journey_conf = {
                    'customer_journey_conf': self.valid_yml
                }
                kwargs['extra_payload'] = customer_journey_conf

            return UssdTestClient(self.live_server_url, **kwargs)



from ussd.core import UssdHandlerAbstract, UssdResponse
from ussd.screens.serializers import UssdContentBaseSerializer
from ussd import defaults as ussd_airflow_variables

class QuitScreen(UssdHandlerAbstract):
    """
    This is the last screen to be shown in a ussd session.

    Its the easiest screen to create. It requires only text

    Example of quit screen:

        .. literalinclude:: .././ussd/tests/sample_screen_definition/valid_quit_screen_conf.yml
    """
    screen_type = "quit_screen"
    serializer = UssdContentBaseSerializer

    def handle(self):
        # set session has expired
        self.ussd_request.session[ussd_airflow_variables.expiry] = True

        if self.initial_screen.get('ussd_report_session'):
            # schedule a task to report session
            self.fire_ussd_report_session_task(self.initial_screen,
                                               self.ussd_request.session_id,
                                               support_countdown=False)

        return UssdResponse(self.get_text(), status=False)



from ussd.core import register_filter
from jinja2.runtime import Undefined


@register_filter
def update(dict_a, dict_b):
    if isinstance(dict_a, Undefined) or dict_a is None:
        dict_a = {}
    dict_a.update(dict_b)
    return dict_a


"""
This test quit screen
"""
from ussd.tests import UssdTestCase


class TestQuitHandler(UssdTestCase.BaseUssdTestCase):
    validation_error_message = dict(
        example_of_quit_screen=dict(
            text=['This field is required.']
        )
    )

    def test(self):
        ussd_client = self.ussd_client(service_code="test_")

        self.assertEqual(
            "END Test getting variable from os environmen. variable_test",
            ussd_client.send('')
        )



from django.apps import AppConfig
import pkgutil



def get_all_screens_module(path_name, package_name, package_list=None):
    package_list = [] if package_list is None else package_list
    for module_loader, name, ispkg in pkgutil.iter_modules((path_name, )):
        if "_screen" in name and not ispkg and\
                        package_name not in package_list:
            __import__(package_name + '.' + name)
            package_list.append(package_name + '.' + name)
    return package_list


class UssdConfig(AppConfig):
    name = 'ussd'

    def ready(self):
        from ussd import screens
        path_name = screens.__path__[0]
        package_name = screens.__name__

        loaded_screens = get_all_screens_module(path_name, package_name)

        for i in loaded_screens:
            print("loaded screens:", i)
        print("loading filters")
        from ussd.filters import date_filters
        from ussd.filters import dict_filters
        from ussd.filters import list_filters

        from ussd.core import env, _registered_filters
        env.filters.update(_registered_filters)



"""
This module is involved in testing Menu screen only
"""
from ussd.tests import UssdTestCase
from ussd.core import ussd_session
from collections import OrderedDict


class TestMenuHandler(UssdTestCase.BaseUssdTestCase):
    choose_meal = "Choose your favourite meal\n" \
                  "1. food\n" \
                  "2. fruits\n" \
                  "3. drinks\n" \
                  "4. vegetables\n" \
                  "5. test pagination\n"

    types_of_food = "Choose your favourite food\n" \
                    "1. rice\n" \
                    "2. back\n"

    types_of_fruit = "No fruits available choose * to go back\n" \
                     "*. back\n"

    rice_chosen = "Your rice will be delivered shortly. " \
                  "Choose 1 to go back\n1. back\n"

    type_of_drinks = "No drinks available choose 0 to go back\n" \
                     "0 back\n"

    error_message = "Please enter a valid choice.\n"

    types_of_vegetables = "Choose one of the following vegetables\n" \
                          "1. Vege Sukuma\n" \
                          "2. Vege Carrot\n" \
                          "3. Vege Cabbage\n"

    choose_quantity = "Choose vegetable size\n" \
                      "1. small at Ksh 50\n" \
                      "2. medium at Ksh 100\n" \
                      "3. large at Ksh 150\n" \
                      "4. back\n"

    selected_vegetables = "You have selected this {vegetable} and " \
                          "this quantity {quantity} at {price}\n" \
                          "1. test_list\n"

    validation_error_message = dict(
        choose_meal=dict(
            non_field_errors=[
                'options field is required or items is required'
            ]
        ),
        types_of_food=dict(
            text=['This field is required.']
        ),
        types_of_fruit=dict(
            options=dict(
                next_screen=["invalid_screen is missing in ussd journey"]
            )
        ),
        types_of_drinks=dict(
            options=dict(
                next_screen=['This field is required.']
            )
        ),
        rice_chosen=dict(
            options=dict(
                next_screen=['This field is required.']
            )
        ),
        types_of_vegetables=dict(
            items=dict(
                value=['This field is required.'],
                session_key=['This field is required.'],
                next_screen=['This field is required.'],
                with_items=['with_items or with_dict field is required'],
                with_dict=['with_items or with_dict field is required']
            )
        )
    )

    @staticmethod
    def add_vegetable_list_in_session(ussd_client):
        session = ussd_session(ussd_client.session_id)
        session["vegetables_list"] = [
            "Sukuma",
            "Carrot",
            "Cabbage"
        ]
        session['vegetable_quantity'] = OrderedDict()
        session['vegetable_quantity']['small'] = 50
        session['vegetable_quantity']['medium'] = 100
        session['vegetable_quantity']['large'] = 150

        session.save()

    def test_happy_case(self):
        ussd_client = self.ussd_client()

        # dial in
        response = ussd_client.send('')

        self.assertEqual(self.choose_meal, response)

        # choose food
        response = ussd_client.send('1')
        self.assertEqual(self.types_of_food, response)

        # choose rice
        response = ussd_client.send('1')
        self.assertEqual(self.rice_chosen, response)

        # choose one to go back
        self.assertEqual(self.choose_meal, ussd_client.send('1'))

        # choose fruits
        self.assertEqual(self.types_of_fruit, ussd_client.send('2'))

        # choose * to go back
        self.assertEqual(self.choose_meal, ussd_client.send('*'))

        # choose 3 for drinks
        self.assertEqual(self.type_of_drinks, ussd_client.send('3'))

        # choose 0 to go back
        self.assertEqual(self.choose_meal, ussd_client.send('0'))

    # todo test invalid option with screen that have multiple pages
    def test_invalid_input(self):
        ussd_client = self.ussd_client()

        # dial in
        ussd_client.send('')

        # select invalid option
        self.assertEqual(ussd_client.send('9'),
                         "You have selected invalid option try again\n" +
                         self.choose_meal
                         )

        # select the correct option 2 for drinks
        self.assertEqual(ussd_client.send('3'), self.type_of_drinks)

        # select invalid back option
        self.assertEqual(ussd_client.send('*'),
                         self.error_message + self.type_of_drinks)

        # select the correct back option
        self.assertEqual(ussd_client.send('0'), self.choose_meal)

    def test_list_screen(self):
        ussd_client = self.ussd_client()
        self.add_vegetable_list_in_session(ussd_client)

        # dial in
        ussd_client.send('')

        # choose vegetables
        self.assertEqual(
            self.types_of_vegetables,
            ussd_client.send('4')
        )

        # choose sukuma
        self.assertEqual(
            self.choose_quantity,
            ussd_client.send('1')
        )

        # choose medium
        self.assertEqual(
            self.selected_vegetables.format(
                vegetable="Sukuma",
                quantity="medium",
                price=100
            ),
            ussd_client.send('2')
        )

        # choose 1 to test native list
        self.assertEqual(
            "1. a\n"
            "2. b\n"
            "3. c\n"
            "4. d\n",
            ussd_client.send('1')
        )

        # choose one ot test native dict items
        response = ussd_client.send('1')
        for i in (
            "a for apple\n",
            "b for boy\n",
            "c for cat\n"
        ):
            self.assertRegex(response, i)

        self.assertEqual(
            "Choose one of the following vegetables\n",
            ussd_client.send('1')
        )

    def test_invalid_input_in_list_screen(self):
        ussd_client = self.ussd_client()

        # dial in
        ussd_client.send('')
        self.add_vegetable_list_in_session(ussd_client)
        # choose vegetables
        ussd_client.send('4')

        # choose invalid option for vegetables
        self.assertEqual(
            self.error_message + self.types_of_vegetables,
            ussd_client.send('9')
        )

        # choose the cabbage
        ussd_client.send('3')

    def test_text_pagination(self):
        ussd_client = self.ussd_client()

        # dial in
        ussd_client.send('')

        # select testing pagination
        response = ussd_client.send('5')

        page_one = "Ussd airflow should be able to wrap anytext " \
                   "that is larger than the one\n98. More\n"

        self.assertEqual(
            response,
            page_one
        )

        # select more
        response = ussd_client.send('98')

        self.assertEqual(
            "specified into two screens.\n1. next\n00. Back\n",
            response
        )

        # test back option
        self.assertEqual(
            page_one,
            ussd_client.send('00')
        )

        self.assertEqual(
            "An example of screen with "
            "multiple options that need to be paginated\n98. More\n",
            ussd_client.send('1')  # select next screen without viewing that option
        )

        # select more to see the next option
        self.assertEqual(
            "1. screen_with_both_text_and_menu_options_pagination\n"
            "00. Back\n"
            "98. More\n",
            ussd_client.send('98')
        )

        # select more to view the next screen
        self.assertEqual(
            "2. screen_with_both_text_item_options_pagination\n00. Back\n",
            ussd_client.send('98')
        )

        # select option 1 to test pagination in both text and options
        self.assertEqual(
            "This screen has both large text and options that exceed the "
            "limit required so\n"
            "98. More\n",
            ussd_client.send('1')
        )

        # select 98 to view more
        self.assertEqual(
            "both the prompt and options will be paginated.\n"
            "1. go back to the previous screen\n"
            "00. Back\n"
            "98. More\n",
            ussd_client.send('98')
        )

        # select 98 to view more
        self.assertEqual(
            "2. quit this session\n"
            "3. this options will be showed in the next_screen\n"
            "00. Back\n",
            ussd_client.send('98')
        )

        # select 3 to test pagination with options and items
        self.assertEqual(
            "This screen has both large text, options, "
            "items that exceed ussd text limit part\n"
            "98. More\n",
            ussd_client.send('3')
        )

        # Todo fix this tests
        # # select 98 to view more
        # self.assertEqual(
        #     "of this text would be displayed in the next screen\n"
        #     "1. apple\n"
        #     "2. boy\n"
        #     "00. Back\n"
        #     "98. More\n",
        #     ussd_client.send('98')
        # )
        #
        # # select 98 to view more
        # self.assertEqual(
        #     "3. cat\n"
        #     "4. dog\n"
        #     "5. egg\n"
        #     "6. frog\n"
        #     "7. girl\n"
        #     "8. house\n"
        #     "9. ice\n"
        #     "10. joyce\n"
        #     "00. Back\n"
        #     "98. More\n",
        #     ussd_client.send('98')
        # )
        #
        # # select 98 to view more
        # self.assertEqual(
        #     "11. kettle\n"
        #     "12. lamp\n"
        #     "13. mum\n"
        #     "14. nurse\n"
        #     "15. ostrich\n"
        #     "16. pigeon\n"
        #     "17. queen\n"
        #     "00. Back\n"
        #     "98. More\n",
        #     ussd_client.send('98')
        # )
        #
        # # select 98 to view more
        # self.assertEqual(
        #     "18. river\n"
        #     "19. sweet\n"
        #     "20. tiger\n"
        #     "21. umbrella\n"
        #     "22. van\n"
        #     "23. water\n"
        #     "24. quit_session\n"
        #     "00. Back\n",
        #     ussd_client.send('98')
        # )
        #
        # # choose apple
        # self.assertEqual(
        #     "end of session apple",
        #     ussd_client.send('1')
        # )


from ussd.tests import UssdTestCase, TestCase
from unittest import mock
from django.test import override_settings
from django.http.response import JsonResponse
from ussd.core import ussd_session
from ussd.tasks import report_session
from uuid import uuid4
from celery.exceptions import MaxRetriesExceededError


@override_settings(
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    CELERY_ALWAYS_EAGER=True,
    BROKER_BACKEND='memory'
)
class TestingUssdReportSession(UssdTestCase.BaseUssdTestCase, TestCase):
    customer_journey_to_use = 'sample_report_session.yml'

    def setUp(self):
        super(TestingUssdReportSession, self).setUp()
        self.valid_yml = self.customer_journey_to_use

    def get_ussd_client(self, journey=customer_journey_to_use):
        return self.ussd_client(
            generate_customer_journey=False,
            extra_payload={
                "customer_journey_conf": journey
            }
        )

    @mock.patch("ussd.core.report_session.apply_async")
    def test_report_session_task(self, report_mock):
        ussd_client = self.get_ussd_client()

        self.assertEqual(
            "Enter you name\n",
            ussd_client.send('')  # dial in
        )

        report_mock.assert_called_once_with(
            args=(ussd_client.session_id,),
            kwargs=dict(
                screen_content={
                    "type": "initial_screen",
                    "next_screen": "screen_one",
                    "ussd_report_session": {
                        "session_key": "reported",
                        "validate_response": [
                            {"expression": "{{reported.status_code}} == 200"}
                        ],
                        "retry_mechanism": {
                            "max_retries": 3
                        },
                        "request_conf": {
                            "url": "localhost:8006/api",
                            "method": "post",
                            "data": {
                                "ussd_interaction": "{{ussd_interaction}}",
                                "session_id": "{{session_id}}"
                            }
                        },
                        "async_parameters": {
                            "queue": "report_session",
                            "countdown": 900
                        }
                    }
                }
            ),
            queue="report_session",
            countdown=900
        )

    @mock.patch("ussd.core.report_session.apply_async")
    def test_quit_screen_sends_report_session(self, mock_report_session):
        ussd_client = self.get_ussd_client()

        self.assertEqual(
            "Enter you name\n",
            ussd_client.send('')  # dial in
        )

        self.assertEqual(
            "Your name is test",
            ussd_client.send('test')  # enter name as test.
        )

        self.assertEqual(mock_report_session.call_count, 2)

        expected_calls = [
            mock.call(
                args=(ussd_client.session_id,),
                kwargs=dict(
                    screen_content={
                        "type": "initial_screen",
                        "next_screen": "screen_one",
                        "ussd_report_session": {
                            "session_key": "reported",
                            "validate_response": [
                                {"expression":
                                    "{{reported.status_code}} == 200"}
                            ],
                            "retry_mechanism": {
                                "max_retries": 3
                            },
                            "request_conf": {
                                "url": "localhost:8006/api",
                                "method": "post",
                                "data": {
                                    "ussd_interaction": "{{ussd_interaction}}",
                                    "session_id": "{{session_id}}"
                                }
                            },
                            "async_parameters": {
                                "queue": "report_session",
                                "countdown": 900
                            }
                        }
                    }
                ),
                queue="report_session",
                countdown=900
            ),
            mock.call(
                args=(ussd_client.session_id,),
                kwargs=dict(
                    screen_content={
                        "type": "initial_screen",
                        "next_screen": "screen_one",
                        "ussd_report_session": {
                            "session_key": "reported",
                            "validate_response": [
                                {"expression":
                                     "{{reported.status_code}} == 200"}
                            ]
                            ,
                            "retry_mechanism": {
                                "max_retries": 3
                            },
                            "request_conf": {
                                "url": "localhost:8006/api",
                                "method": "post",
                                "data": {
                                    "ussd_interaction": "{{ussd_interaction}}",
                                    "session_id": "{{session_id}}"
                                }
                            },
                            "async_parameters": {
                                "queue": "report_session",
                                "countdown": 900
                            }
                        }
                    }
                ),
                queue="report_session",
            )
        ]

        mock_report_session.assert_has_calls(expected_calls)

    @staticmethod
    @mock.patch('ussd.core.requests.request')
    def test_http_call(mock_request):
        mock_response = JsonResponse({"balance": 250})
        mock_request.return_value = mock_response

        session = ussd_session(str(uuid4()))
        session['session_id'] = session.session_key
        session['ussd_interaction'] = []
        session.save()

        screen_content = {
            "type": "initial_screen",
            "next_screen": "screen_one",
            "ussd_report_session": {
                "session_key": "reported",
                "retry_mechanism": {
                    "max_retries": 3
                },
                "validate_response": [
                                {"expression":
                                    "{{reported.status_code}} == 200"}
                            ],
                "request_conf": {
                    "url": "localhost:8006/api",
                    "method": "post",
                    "data": {
                        "ussd_interaction": "{{ussd_interaction}}",
                        "session_id": "{{session_id}}"
                    }
                },
                "async_parameters": {
                    "queue": "report_session",
                    "countdown": 900
                }
            }
        }

        report_session.delay(
            session.session_key,
            screen_content
        )

        mock_request.assert_called_once_with(
            url="localhost:8006/api",
            method="post",
            data=dict(
                ussd_interaction='[]',
                session_id=session.session_key
            )
        )

    @mock.patch("ussd.tasks.requests.request")
    def test_if_session_is_already_posted_wont_post_again(self, mock_request):
        mock_response = JsonResponse({"balance": 250})
        mock_request.return_value = mock_response

        session = ussd_session(str(uuid4()))
        session['session_id'] = session.session_key
        session['ussd_interaction'] = []
        session['posted'] = True
        session.save()

        screen_content = {
            "type": "initial_screen",
            "next_screen": "screen_one",
            "ussd_report_session": {
                "session_key": "reported",
                "retry_mechanism": {
                    "max_retries": 3
                },
                "validate_response": [
                                {"expression":
                                    "{{reported.status_code}} == 200"}
                            ],
                "request_conf": {
                    "url": "localhost:8006/api",
                    "method": "post",
                    "data": {
                        "ussd_interaction": "{{ussd_interaction}}",
                        "session_id": "{{session_id}}"
                    }
                },
                "async_parameters": {
                    "queue": "report_session",
                    "countdown": 900
                }
            }
        }

        report_session.delay(
            session.session_key,
            screen_content
        )
        self.assertFalse(mock_request.called)

    @mock.patch("requests.request")
    @mock.patch.object(report_session, 'retry')
    def test_retry(self, mock_retry, mock_request):
        mock_response = JsonResponse({"balance": 250},
                                     status=400
                                     )
        mock_request.return_value = mock_response
        ussd_client = self.get_ussd_client()
        ussd_client.send('mwas')

        # check posted flag has been set to false
        self.assertFalse(
            ussd_session(ussd_client.session_id)['posted'],
        )

        self.assertTrue(mock_retry.called)

    @mock.patch("ussd.core.report_session.apply_async")
    def test_report_task_only_called_when_activated(self, mock_report_session):
        # using a journey that report_session has not been activated
        # and one that has quit screen
        ussd_client = self.get_ussd_client(
            journey='valid_quit_screen_conf.yml')

        self.assertEqual(
            "Test getting variable from os environmen. variable_test",
            ussd_client.send('') # dial in
        )

    @mock.patch("ussd.core.requests.request")
    @mock.patch.object(report_session, 'retry')
    def test_maximum_retries(self, mock_retry, mock_request):
        mock_response = JsonResponse({"balance": 250},
                                     status=400
                                     )
        mock_request.return_value = mock_response
        mock_retry.side_effect = MaxRetriesExceededError()

        ussd_client = self.get_ussd_client()
        ussd_client.send('mwas')


    def testing_invalid_customer_journey(self):
        # this is tested in the initial screen
        pass





from ussd.core import UssdView, UssdRequest
from ussd.tests.sample_screen_definition import path
from django.http import HttpResponse
from django.conf import settings


class AfricasTalkingUssdGateway(UssdView):

    def post(self, req):
        list_of_inputs = req.data['text'].split("*")
        text = "*" if len(list_of_inputs) >= 2 and list_of_inputs[-1] == "" and list_of_inputs[-2] == "" else list_of_inputs[
            -1]

        session_id = req.data['sessionId']
        if req.data.get('use_built_in_session_management', False):
            session_id = None
        ussd_request = UssdRequest(
            phone_number=req.data['phoneNumber'].strip('+'),
            session_id=session_id,
            ussd_input=text,
            service_code=req.data['serviceCode'],
            language=req.data.get('language', 'en'),
            use_built_in_session_management=req.data.get(
                'use_built_in_session_management', False)
        )

        return ussd_request

    def get_customer_journey_conf(self, request):
        if request.data.get('customer_journey_conf'):
            return path + '/' + request.data.get('customer_journey_conf')
        return getattr(settings, 'DEFAULT_USSD_SCREEN_JOURNEY',
                       path + "/sample_customer_journey.yml")

    def get_customer_journey_namespace(self, request):
        if request.data.get('customer_journey_conf'):
            return request.data['customer_journey_conf'].replace(
                '.yml', ''
            )
        return "AfricasTalkingUssdGateway"

    def ussd_response_handler(self, ussd_response):
        if self.request.data.get('serviceCode') == 'test':
            return super(AfricasTalkingUssdGateway, self).\
                ussd_response_handler(ussd_response)
        if ussd_response.status:
            res = 'CON' + ' ' + str(ussd_response)
            response = HttpResponse(res)
        else:
            res = 'END' + ' ' + str(ussd_response)
            response = HttpResponse(res)
        return response



from ussd.core import UssdHandlerAbstract, UssdHandlerMetaClass
from ussd.screens.serializers import UssdBaseSerializer
from rest_framework import serializers
from ussd.utilities import str_to_class


class CustomScreenSerializer(UssdBaseSerializer):
    screen_obj = serializers.CharField(max_length=255)

    @staticmethod
    def validate_screen_obj(value):
        try:
            screen_obj = str_to_class(value)
        except Exception as e:
            raise serializers.ValidationError(str(e))
        else:
            if not isinstance(screen_obj, UssdHandlerMetaClass):
                raise serializers.ValidationError(
                    "Screen object should be of type UssdHandlerAbstract"
                )
            return screen_obj

    def validate(self, data):
        data = super(CustomScreenSerializer, self).validate(data)
        screen_obj = data['screen_obj']
        if hasattr(screen_obj, 'serializer'):
            validation = screen_obj.serializer(data=self.initial_data,
                                               context=self.context)
            if not validation.is_valid():
                raise serializers.ValidationError(validation.errors)
        return data


class CustomScreen(UssdHandlerAbstract):
    """
    If you have a particular user case that's not yet covered by
    our existing screens, this is the screen to use.

    This screen allows us to define our own ussd screen.

    To create it you need the following fields.
        1. screen_object
            This is the path to be used to import the class
        2. serializer (optional)
            This if you want to be validating your screen with
            specific fields
        3. You can define any field that you feel
            your custom screen might need.
    EXAMPLE:
        examples of custom screen 
            .. code-block:: python
            
                class SampleCustomHandler1(UssdHandlerAbstract):
                    abstract = True  # don't register custom classes
                    @staticmethod
                    def show_ussd_content():  # This method doesn't have to be static
                        # Do anything custom here.
                        return UssdResponse("This is a custom Handler1")
                
                    def handle_ussd_input(self, ussd_input):
                        # Do anything custom here
                        print(ussd_input)  # pep 8 for the sake of using it.
                        return self.ussd_request.forward('custom_screen_2')
    
    
                class SampleSerializer(UssdBaseSerializer, NextUssdScreenSerializer):
                        input_identifier = serializers.CharField(max_length=100)
                
                
                class SampleCustomHandlerWithSerializer(UssdHandlerAbstract):
                    abstract = True  # don't register custom classes
                    serializer = SampleSerializer
                
                    @staticmethod
                    def show_ussd_content():  # This method doesn't have to be static
                        return "Enter a digit and it will be doubled on your behalf"
                
                    def handle_ussd_input(self, ussd_input):
                        self.ussd_request.session[
                            self.screen_content['input_identifier']
                        ] = int(ussd_input) * 2
                
                        return self.ussd_request.forward(
                            self.screen_content['next_screen']
                        )
        
        example of defining a yaml
            
            .. literalinclude:: .././ussd/tests/sample_screen_definition/valid_menu_screen_conf.yml
    """
    screen_type = "custom_screen"
    serializer = CustomScreenSerializer

    def handle(self):
        # Todo initiate the class in the core. to avoid double changing
        # if the parameter changes.

        # calling the custom screen handler method
        return str_to_class(
            self.screen_content['screen_obj']
        )(
            self.ussd_request,
            self.handler,
            self.screen_content,
            initial_screen={},
        ).handle()



from django.core.management import call_command
from django.test import TestCase
from django.utils.six import StringIO
from .sample_screen_definition import path
from django.core.management.base import CommandError
import json


class ValidateCustomerJourneyConfig(TestCase):
    maxDiff = None

    def test_command_output(self):
        out = StringIO()
        file_name = "{0}/valid_quit_screen_conf.yml".format(path)
        call_command('validate_ussd_journey', file_name, stdout=out)
        expected_output = {
            file_name: dict(
                valid=True,
                error_message=dict()
            )
        }
        self.assertDictEqual(expected_output, json.loads(out.getvalue()))

    def testing_invalid_ussd_journey(self):
        out = StringIO()
        file_name = "{0}/invalid_quit_screen_conf.yml".format(path)
        self.assertRaises(CommandError, call_command, 'validate_ussd_journey',
                          file_name, stdout=out)

    def test_called_with_multiple_files(self):
        out = StringIO()
        file_1 = "{0}/valid_quit_screen_conf.yml".format(path)
        file_2 = "{0}/valid_http_screen_conf.yml".format(path)
        file_3 = "{0}/valid_input_screen_conf.yml".format(path)
        call_command(
            "validate_ussd_journey",
            file_1,
            file_2,
            file_3,
            stdout=out)

        expected_output = {
            file_1: dict(
                valid=True,
                error_message={},
            ),
            file_2: dict(
                valid=True,
                error_message={}
            ),
            file_3: dict(
                valid=True,
                error_message={}
            )
        }

        self.assertDictEqual(expected_output, json.loads(out.getvalue()))

    def test_called_with_invalid_file_path(self):
        out = StringIO()

        self.assertRaises(CommandError, call_command, 'validate_ussd_journey', 'invalid_path', stdout=out)


from ussd.core import UssdHandlerAbstract
from ussd.screens.serializers import UssdBaseSerializer, \
    NextUssdScreenSerializer
from rest_framework import serializers


class RouterOptionSerializer(NextUssdScreenSerializer):
    expression = serializers.CharField(max_length=255)


class RouterSerializer(UssdBaseSerializer):
    router_options = serializers.ListField(
        child=RouterOptionSerializer()
    )


class RouterScreen(UssdHandlerAbstract):
    """
    This screen is invisible to the user. Sometimes you would like to
    direct user to different screens depending on some status.

    For instance you want to show different screen to users who are not
    registered and a different screen to users who have already registered.
    This is the screen to create.

    Fields used to create this screen:
        1. router_options
            This is a list of router option.
            Each router option has the following fields
                a. expression
                    This is a jinja expression that's is evaluating to boolean
                    It can reference anything in the session and parameters
                    in ussd_request
                b. next_screen
                    This is the screen to direct to if the above expression
                    is true
        2. default_next_screen (optional)
            This is the screen to direct to if all expression in router_options
            failed.

        3. with_items (optional)
            Sometimes you want to loop over something until an item
            passes the expression. In this case use with_items.
            When using with_items you can use variable item in the
            expression.

            see in the example below for more explanation

        Examples of router screens

            .. literalinclude:: .././ussd/tests/sample_screen_definition/valid_router_screen_conf.yml
    """

    screen_type = "router_screen"
    serializer = RouterSerializer

    def handle(self):
        route_options = self.screen_content.get("router_options")
        loop_items = [0]
        if self.screen_content.get("with_items"):
            loop_items = self.evaluate_jija_expression(
                self.screen_content["with_items"],
                session=self.ussd_request.session
            ) or loop_items

        for item in loop_items:
            extra_context = {
                "item": item
            }
            if isinstance(loop_items, dict):
                extra_context.update(
                    dict(
                        key=item,
                        value=loop_items[item],
                        item={item: loop_items[item]}
                    )
                )

            for option in route_options:
                if self.evaluate_jija_expression(
                        option['expression'],
                        session=self.ussd_request.session,
                        extra_context=extra_context
                ):
                    return self.ussd_request.forward(option['next_screen'])
        return self.ussd_request.forward(
            self.screen_content['default_next_screen']
        )



from ussd.core import UssdHandlerAbstract, UssdResponse
from ussd.screens.serializers import UssdContentBaseSerializer, \
    UssdTextSerializer, NextUssdScreenSerializer, MenuOptionSerializer
from django.utils.encoding import force_text
import re
from rest_framework import serializers
from ussd.screens.menu_screen import MenuScreen


class InputValidatorSerializer(UssdTextSerializer):
    regex = serializers.CharField(max_length=255, required=False)
    expression = serializers.CharField(max_length=255, required=False)

    def validate(self, data):
        return super(InputValidatorSerializer, self).validate(data)


class InputSerializer(UssdContentBaseSerializer, NextUssdScreenSerializer):
    input_identifier = serializers.CharField(max_length=100)
    validators = serializers.ListField(
        child=InputValidatorSerializer(),
        required=False
    )
    options = serializers.ListField(
        child=MenuOptionSerializer(),
        required=False
    )


class InputScreen(MenuScreen):
    """

    This screen prompts the user to enter an input

    Fields required:
        - text: this the text to display to the user.
        - input_identifier: input amount entered by users will be saved
                            with this key. To access this in the input
                            anywhere {{ input_identifier }}
        - next_screen: The next screen to go after the user enters
                        input
        - validators:
            - text: This is the message to display when the validation fails
              regex: regex used to validate ussd input. Its mutually exclusive
              with expression
            - expression: if regex is not enough you can use a jinja expression
             will be called ussd request object
              text: This the message thats going to be displayed if expression
              returns False
        - options (This field is optional):
            This is a list of options to display to the user
            each option is a key value pair of option text to display
            and next_screen to redirect if option is selected.
            Example of option:

            .. code-block:: yaml

                   options:
                    - text: option one
                      next_screen: screen_one
                    - text: option two
                      next_screen: screen_two

    Example:
        .. literalinclude:: .././ussd/tests/sample_screen_definition/valid_input_screen_conf.yml
    """

    screen_type = "input_screen"
    serializer = InputSerializer

    def handle_invalid_input(self):
        # validate input
        validation_rules = self.screen_content.get("validators", {})
        for validation_rule in validation_rules:

            if 'regex' in validation_rule:
                regex_expression = validation_rule['regex']
                regex = re.compile(regex_expression)
                is_valid = bool(
                    regex.search(
                        force_text(self.ussd_request.input)
                    ))
            else:
                is_valid = self.evaluate_jija_expression(
                    validation_rule['expression'],
                    session=self.ussd_request.session
                )

            # show error message if validation failed
            if not is_valid:
                return UssdResponse(
                    self.get_text(
                        validation_rule['text']
                    )
                )

        session_key = self.screen_content['input_identifier']
        next_handler = self.screen_content['next_screen']
        self.ussd_request.session[session_key] = \
            self.ussd_request.input

        return self.ussd_request.forward(next_handler)



from django.test import TestCase
from ussd.core import _registered_ussd_handlers, \
    UssdHandlerAbstract, MissingAttribute, \
    InvalidAttribute, UssdRequest, ussd_session
from rest_framework import serializers
from ussd.tests import UssdTestCase
from freezegun import freeze_time
from datetime import datetime
import time
from ussd import defaults as ussd_airflow_variables
from ussd.utilities import datetime_to_string, string_to_datetime


class SampleSerializer(serializers.Serializer):
    text = serializers.CharField()


class TestHandlerRegistration(TestCase):

    def test_happy_case(self):

        class TestOne(UssdHandlerAbstract):
            screen_type = "test_one"
            serializer = SampleSerializer

            def handle(self, req):
                pass

        self.assertTrue(_registered_ussd_handlers.get("test_one"))

        self.assertTrue(
            _registered_ussd_handlers['test_one'] == TestOne
        )

    @staticmethod
    def test_missing_screen_type_attribute():

        try:
            # missing screen_type
            class TestTwo(UssdHandlerAbstract):
                serializer = SampleSerializer
                def handle(self, req):
                    pass

            assert False, "should raise missing attriute name"
        except MissingAttribute:
            pass

    @staticmethod
    def test_missing_serializer_attribute():

        try:
            # missing validate schema
            class TestFour(UssdHandlerAbstract):
                screen_type = 'test_four'


            assert False, "should raise missing attriute name"
        except MissingAttribute:
            pass

    @staticmethod
    def test_invalid_serializer():

        try:
            # invalid serializer
            class TestFive(UssdHandlerAbstract):
                screen_type = 'test_five'
                serializer = "Sample serializer"

                def handle(self, req):
                    pass


            assert False, "should raise invalid serializer"
        except InvalidAttribute:
            pass


class TestUssdRequestCreation(TestCase):

    def test(self):
        session_id = '1234'
        ussd_request = UssdRequest('1234', '200', '', 'en')

        self.assertTrue(len(session_id) < 8)

        self.assertTrue(len(ussd_request.session_id) >= 8)

        # test creating with session id as None
        self.assertRaises(InvalidAttribute,
                          UssdRequest,
                          None,
                          '200',
                          '',
                          'en')


        # session_id and using custom session_id can't been defined
        self.assertRaises(
            InvalidAttribute,
            UssdRequest,
            "sfewevae",
            "200",
            "",
            "en",
            use_built_in_session_management=True
        )


class TestCoreView(UssdTestCase.BaseUssdTestCase):
    validate_ussd = False

    def test_africas_talking_is_picking_settings_journey(self):
        ussd_client = self.ussd_client(generate_customer_journey=False)

        # dial in
        response = ussd_client.send('')

        self.assertEqual(
            "Enter your name\n",
            response
        )


class TestInheritance(UssdTestCase.BaseUssdTestCase):

    def get_client(self):
        return self.ussd_client(
            generate_customer_journey=False,
            extra_payload={
                "customer_journey_conf": "sample_using_inheritance.yml"
            }
        )

    def test(self):
        ussd_client = self.get_client()

        inherited_text = "Enter anything\n"

        # dial in
        self.assertEqual(
            inherited_text,
            ussd_client.send('')
        )

        # enter my first name
        self.assertEqual(
            inherited_text,
            ussd_client.send('Francis')
        )

        # enter second name
        self.assertEqual(
            "First input was Francis and second input was Mwangi\n"
            "1. Continue\n",
            ussd_client.send("Mwangi")
        )

    @freeze_time(datetime.now())
    def test_steps_recording(self):
        ussd_client = self.get_client()

        # dial in
        ussd_client.send('')

        # enter first name
        ussd_client.send('Francis')

        # enter second name
        ussd_client.send('Mwangi')

        # enter 1 to continue
        ussd_client.send('1')

        # press two to go back
        ussd_client.send('2')

        # enter 1 to continue
        ussd_client.send('1')

        # enter 1 to exit
        self.assertEqual(
            "This is the last screen",
            ussd_client.send('1')
        )

        now = datetime_to_string(datetime.now())
        expected_screen_interaction = [
            {
                "screen_name": "screen_one",
                "screen_text": "Enter anything\n",
                "input": "Francis",
                "start_time": now,
                "end_time": now,
                "duration": 0.0
            },
            {
                "screen_name": "screen_two",
                "screen_text": "Enter anything\n",
                "input": "Mwangi",
                "start_time": now,
                "end_time": now,
                "duration": 0.0
            },
            {
                "screen_name": "screen_three",
                "screen_text": "First input was Francis and "
                               "second input was Mwangi\n1. Continue\n",
                "input": "1",
                "start_time": now,
                "end_time": now,
                "duration": 0.0
            },
            {
                "screen_name": "screen_four",
                "screen_text": "Press 1 to exit or 2 to go back\n"
                               "1. Exit\n"
                               "2. Back\n",
                "input": "2",
                "start_time": now,
                "end_time": now,
                "duration": 0.0
            },
            {
                "screen_name": "screen_three",
                "screen_text": "First input was Francis and "
                               "second input was Mwangi\n1. Continue\n",
                "input": "1",
                "start_time": now,
                "end_time": now,
                "duration": 0.0
            },
            {
                "screen_name": "screen_four",
                "screen_text": "Press 1 to exit or 2 to go back\n"
                               "1. Exit\n"
                               "2. Back\n",
                "input": "1",
                "start_time": now,
                "end_time": now,
                "duration": 0.0
            },
            {
                "screen_name": "screen_five",
                "screen_text": "This is the last screen",
                "input": "",
                "start_time": now
            }

        ]

        session = ussd_session(ussd_client.session_id)

        self.assertEqual(
            session['ussd_interaction'],
            expected_screen_interaction
        )

    def testing_valid_customer_journey(self):
        self._test_ussd_validation(
            'sample_using_inheritance.yml',
            True,
            {}
        )

    def testing_invalid_customer_journey(self):
        pass


class TestSessionManagement(UssdTestCase.BaseUssdTestCase):
    validate_ussd = False

    def get_client(self):
        return self.ussd_client(
            generate_customer_journey=False,
            extra_payload={
                "customer_journey_conf":
                    "sample_used_for_testing_session_management.yml",
                "use_built_in_session_management": True
            }
        )

    @staticmethod
    def _create_ussd_request(phone_number):
        return UssdRequest(None, phone_number, '', 'en',
                           use_built_in_session_management=True,
                           expiry=1)

    def test_session_expiry(self):
        phone_number = 200
        req = self._create_ussd_request(phone_number)

        # test a new session created within the time session id remains
        self.assertEqual(req.session_id,
                         self._create_ussd_request(phone_number).session_id
                         )
        time.sleep(1)

        # test after one second a new session_id will be created.
        self.assertNotEqual(req.session_id,
                            self._create_ussd_request(phone_number).session_id
                            )

    def test_session_expiry_using_inactivity(self):
        # Test session expiry is using last time session was updated
        phone_number = '201'
        req = self._create_ussd_request(phone_number)

        time.sleep(0.8)

        # confirm session_id hasn't changed
        self.assertEqual(
            req.session_id,
            self._create_ussd_request(phone_number).session_id
        )

        # make a change in session
        req = self._create_ussd_request(phone_number)
        req.session['name'] = 'mwas'
        req.session[ussd_airflow_variables.last_update] = datetime_to_string(
            datetime.now())
        req.session.save()

        time.sleep(0.8)

        # confirm even after the session was created 1.6 sec ago hasn't
        # been closed

        self.assertEqual(
            req.session_id,
            self._create_ussd_request(phone_number).session_id
        )

        time.sleep(0.3)

        # test now the session has been closed
        self.assertNotEqual(
            req.session_id,
            self._create_ussd_request(phone_number).session_id
        )

    def test_quit_screen_terminates_session(self):

        ussd_client = self.get_client()

        self.assertEqual(
            "Enter your name\n",
            ussd_client.send('')  # dial in
        )

        self.assertEqual(
            "Enter your age\n",
            ussd_client.send('Mwas')  # enter name
        )

        self.assertEqual(
            "You have entered your name as Mwas "
            "and your age as 24",
            ussd_client.send('24')
        )

        # dial in again to test the session has been closed
        self.assertEqual(
            "Enter your name\n",
            ussd_client.send('1')
        )

    def test_session_expiry_on_ussd(self):

        ussd_client = self.get_client()

        self.assertEqual(
            "Enter your name\n",
            ussd_client.send('')
        )

        # sleep for 1 sec for session to expiry
        time.sleep(1)

        self.assertEqual(
            "Enter your name\n",
            ussd_client.send('')
        )


class TestDatetimeConversion(TestCase):

    def test(self):
        now = datetime.now()
        now_str = datetime_to_string(now)

        self.assertEqual(now, string_to_datetime(now_str))



from ussd.core import UssdHandlerAbstract, UssdResponse
from .serializers import UssdContentBaseSerializer, \
    MenuOptionSerializer, NextUssdScreenSerializer, UssdTextSerializer
from rest_framework.serializers import ListField, ValidationError, \
    CharField
from django.core.paginator import Paginator
import textwrap

class WithItemField(CharField):
    def to_internal_value(self, data):
        if isinstance(data, list):
            return data
        return super(WithItemField, self).to_internal_value(data)


class WithDictField(CharField):
    def to_internal_value(self, data):
        if isinstance(data, dict):
            return data
        return super(WithDictField, self).to_internal_value(data)


class ItemsSerializer(UssdTextSerializer, NextUssdScreenSerializer):
    value = CharField()
    session_key = CharField()
    with_items = WithItemField(required=False, default=None)
    with_dict = WithDictField(required=False, default=None)

    def to_internal_value(self, data):
        # set data in instance to access later
        self.custom_data = data
        return super(ItemsSerializer, self).to_internal_value(data)

    def validate_with_items(self, value, loop_value='with_dict'):
        if value is None and self.custom_data.get(loop_value, None) is None:
            raise ValidationError(
                "with_items or with_dict field is required"
            )
        return value

    def validate_with_dict(self, value):
        return self.validate_with_items(value, 'with_items')


class MenuScreenSerializer(UssdContentBaseSerializer):
    """
    - text:
        .. autoclass:: UssdContentBaseSerializer

    - options:
        This is a list of options to display to the user
        each option is
            .. autoclass:: MenuOptionSerializer

    - items:
        Sometimes you what to display a list of items and not
        menu options. Item is used for this
            .. autoclass:: ItemsSerializer

    - error_message: This is an optional message to display if the
                     user enters invalid input

    - option and items are mutual exclusive.

    Examples of menu screen:

        .. literalinclude:: .././ussd/tests/sample_screen_definition/valid_menu_screen_conf.yml
    """
    options = ListField(
        child=MenuOptionSerializer(),
        required=False
    )
    items = ItemsSerializer(required=False)

    def validate(self, data):
        if 'options' not in data and 'items' not in data:
            raise ValidationError(
                'options field is required or items is required'
            )
        return super(MenuScreenSerializer, self).validate(data)


class ListItem(object):
    def __init__(self, text, value):
        self.text = text
        self.value = value


class MenuOption(object):
    def __init__(self, text, next_screen, index_display=None,
                 index_value=None):
        self.text = text
        self.next_screen = next_screen
        self.index_display = index_display or index_value
        self.index_value = index_value or self.index_display


class MenuScreen(UssdHandlerAbstract):
    """
    This is the screen used to display options to select:

        - text:
            This is the text to display to the user.

        - options:
            This is a list of options to display to the user
            each option is a key value pair of option text to display
            and next_screen to redirect if option is selected.
            Example of option:

            .. code-block:: yaml

                   options:
                    - text: option one
                      next_screen: screen_one
                    - text: option two
                      next_screen: screen_two

        - items:
            Unlike options where each option has its own screen to redirect
            in items we have a list of items to display and regardless of
            the input user will be redirected to one screen.

                Example of items

                .. code-block:: yaml

                    menu_screen_with_item_example:
                        type: menu_screen
                        text: choose one item
                        items:
                            text: "{{key}} for {{value}}"
                            value: "{{item}}"
                            next_screen: display_option
                            session_key: testing
                            with_dict:
                                a: apple
                                b: boy
                                c: cat

            In the above example if will display the following text

            .. code-block:: text

                    Choose one item
                        1. apple
                        2. boy
                        3. cat

            If the user selects "2", that would be translated by the value
            key, it will result to "b", then "b" will be saved with session_key
            provided and the user will be directed to the next screen which
            is display_option.

            To reference the selected item, use {{your_session_key}}

        - error_message: (optional)
            This is message to display if the user enter the wrong value.

            defaults to "Please enter a valid choice."

        - option and items are mutual exclusive.

    Example:
        .. literalinclude:: .././ussd/tests/sample_screen_definition/valid_menu_screen_conf.yml
    """
    screen_type = "menu_screen"
    serializer = MenuScreenSerializer

    def __init__(self, *args, **kwargs):
        super(MenuScreen, self).__init__(*args, **kwargs)
        self.list_options = [] if self.screen_content.get('items') is None \
            else self.get_items()
        self.menu_options = [] if self.screen_content.get('options') is None \
            else self.get_menu_options()
        self.error_message = "Please enter a valid choice.\n" \
            if not self.screen_content.get('error_message') \
            else self.get_text(self.screen_content["error_message"])

        # all options
        self.options = self.list_options + \
                       ([] if self.screen_content.get('options') is None else
                        self.get_menu_options(
                            start_index=len(self.list_options) + 1))

        self.paginator = self.get_paginator()

    def show_ussd_content(self):
        self.ussd_request.session['_ussd_state']['page'] = 1
        return self._render_django_page(1)

    def _render_django_page(self, index):
        return self.paginator.page(index).object_list[0]

    def get_paginator(self):
        pages = []
        ussd_title = self._add_end_line(self.get_text())

        # get ussd text limit
        ussd_text_limit = self.get_text_limit()

        # paginate menu screen text
        while len(ussd_title) > ussd_text_limit:
            # Lets create pages
            text = ""
            if len(pages) > 0:
                text += "00. {back_option}".format(
                    back_option=self.pagination_back_option)
            text += "98. {more_option}".format(
                more_option=self.pagination_more_option
            )

            # update ussd_text_limit to the one that considers pages
            ussd_text_limit = ussd_text_limit - len(text) - 1

            ussd_text_subsets = textwrap.wrap(
                ussd_title, width=ussd_text_limit
            )

            pages.append(
                self._add_end_line(ussd_text_subsets[0]) + text
            )

            ussd_title = self._add_end_line(' '.join(ussd_text_subsets[1:]))

        self.paginate_options(ussd_title, pages, self.options[:])

        return Paginator(pages, 1)

    def paginate_options(self, ussd_text, pages, options):
        """
        Assumptions:
            - ussd_text is within the limit
        """
        # Todo use back off strategy to generate the pages
        text = ""
        if len(pages) > 0:
            text += "00. {back_option}".format(
                back_option=self.pagination_back_option)

        if not options:
            pages.append(
                ussd_text + text
            )
            return pages

        ussd_text_cadidate = ussd_text + options[0].text
        # detect if there might be more optoins
        text += "98. {more_option}".format(more_option=
                                           self.pagination_more_option) \
            if len(ussd_text_cadidate) > self.get_text_limit() - len(text) \
            else ''
        if len(ussd_text_cadidate) <= self.get_text_limit() - len(text):
            ussd_text = ussd_text + options[0].text
        else:
            pages.append(
                ussd_text + text
            )
            ussd_text = options[0].text
        return self.paginate_options(
                ussd_text,
                pages,
                options[1:]
            )

    def handle_ussd_input(self, ussd_input):
        # check if input is for previous or next page
        if self.ussd_request.input.strip() in ("98", "00"):
            page = self.paginator.page(
                self.ussd_request.session['_ussd_state']['page']
            )
            if self.ussd_request.input.strip() == "98" and page.has_next():
                new_page_number = page.next_page_number()
                self.ussd_request.session['_ussd_state']['page'] = \
                    new_page_number
                return UssdResponse(self._render_django_page(new_page_number))
            elif self.ussd_request.input.strip() == '00' and \
                    page.has_previous():
                new_page_number = page.previous_page_number()
                self.ussd_request.session['_ussd_state']['page'] = \
                    new_page_number
                return UssdResponse(self._render_django_page(new_page_number))
        next_screen = self.evaluate_input()
        if next_screen:
            return self.ussd_request.forward(next_screen)
        return self.handle_invalid_input()

    def evaluate_input(self):
        """
        This gets the selected option,
        and returns next_screen, and error message if any
        :return:
        """
        if self.ussd_request.input.isdigit() and \
                not int(self.ussd_request.input) <= 0:
            ussd_input = int(self.ussd_request.input)
            ussd_input_index = ussd_input - 1
            if ussd_input <= len(self.list_options):
                # save input in the session
                selected_item = self.list_options[ussd_input_index]
                self.ussd_request.session[
                    self.screen_content['items']['session_key']] = \
                    selected_item.value
                # forward request to the next screen
                return self.screen_content['items']['next_screen']
            elif ussd_input <= len(self.menu_options):
                return self.menu_options[ussd_input_index].next_screen
        else:
            for option in self.menu_options:
                if option.index_value == self.ussd_request.input:
                    return option.next_screen
        return False

    def get_items(self, start_index: int = 1) -> list:
        """
        This gets ListItems
        :return:
        """
        items_section = self.screen_content['items']

        text = self.screen_content['items']['text']
        value = self.screen_content['items']['value']

        loop_method = ""
        loop_value = ""
        for key, value_ in items_section.items():
            if key.startswith("with_"):
                loop_method = "_" + key
                loop_value = value_

        items = self.evaluate_jija_expression(loop_value,
                                              session=self.ussd_request.session
                                              )

        return getattr(self, loop_method)(
            text, value, items, start_index
        )

    def get_menu_options(self, start_index: int = 1) -> list:
        menu_options = []
        for i, option in enumerate(self.screen_content.get('options', []),
                                   start_index):
            input_value = option.get('input_value') or i
            input_display = option.get('input_display') or \
                            "{index}. ".format(index=input_value)
            text = "{display_option}{text}".format(
                display_option=input_display,
                text=self._add_end_line(
                    self.get_text(text_context=option['text'])
                )
            )
            menu_options.append(
                MenuOption(
                    text,
                    option['next_screen'],
                    input_display,
                    input_value
                )
            )
        return menu_options

    def handle_invalid_input(self):
        return UssdResponse(
            self._add_end_line(
                self.get_text(self.error_message)) +
            self._render_django_page(1)
        )

    def _with_items(self, text, value, items, start_index):
        list_items = []
        for index, item in enumerate(items, start_index):
            context = {}
            extra = {
                "item": item
            }
            if isinstance(items, dict):
                extra.update(
                    dict(
                        key=item,
                        value=items[item],
                        item={item: items[item]}
                    )
                )

            context.update(extra)

            list_items.append(
                ListItem(
                    self._add_end_line("{index}. {text}".format(
                        index=index,
                        text=UssdHandlerAbstract.render_text(
                            self.ussd_request.session,
                            text,
                            extra=context
                        )
                    )
                    ),
                    self.evaluate_jija_expression(value,
                                                  session=
                                                  self.ussd_request.session,
                                                  extra_context=context)
                )
            )
        return list_items

    def _with_dict(self, text, value, items, start_index):
        return self._with_items(text, value, items, start_index)





"""
Comming soon
"""
from urllib.parse import unquote
from copy import copy, deepcopy
from rest_framework.views import APIView
from django.http import HttpResponse
from structlog import get_logger
import staticconf
from django.conf import settings
from importlib import import_module
from django.contrib.sessions.backends import signed_cookies
from django.contrib.sessions.backends.base import CreateError
from jinja2 import Template, Environment, TemplateSyntaxError
from .screens.serializers import UssdBaseSerializer
from rest_framework.serializers import SerializerMetaclass
import re
import json
import os
from configure import Configuration
from datetime import datetime
from ussd.models import SessionLookup
from annoying.functions import get_object_or_None
from ussd import defaults as ussd_airflow_variables
from django.utils import timezone
import requests
import inspect
from ussd.tasks import report_session
from ussd import utilities

_registered_ussd_handlers = {}
_registered_filters = {}

# initialize jinja2 environment
env = Environment(keep_trailing_newline=True)
env.filters.update(_registered_filters)


class MissingAttribute(Exception):
    pass


class InvalidAttribute(Exception):
    pass


class DuplicateSessionId(Exception):
    pass


def register_filter(func_name, *args, **kwargs):
    filter_name = func_name.__name__
    _registered_filters[filter_name] = func_name


def get_session_engine():
    session_engine = import_module(getattr(settings, "USSD_SESSION_ENGINE",
                                           settings.SESSION_ENGINE))
    if session_engine is signed_cookies:
        raise ValueError("You cannot use channels session "
                         "functionality with signed cookie sessions!")
    return session_engine


def ussd_session(session_id):
    session = get_session_engine().SessionStore(session_key=session_id)
    session._session.keys()
    session._session_key = session_id

    # If the session does not already exist, save to force our
    # session key to be valid.
    if not session.exists(session.session_key):
        try:
            session.save(must_create=True)
        except CreateError:
            # Session wasn't unique, so another consumer is doing the same thing
            raise DuplicateSessionId("another sever is working"
                                     "on this session id")
    return session


def generate_session_id():
    session_store = get_session_engine().SessionStore()
    session_store.save()  # generate session_key
    return session_store.session_key


def load_yaml(file_path, namespace):
    file_path = Template(file_path).render(os.environ)
    yaml_dict = Configuration.from_file(
            os.path.abspath(file_path),
            configure=False
        )
    staticconf.DictConfiguration(
        yaml_dict,
        namespace=namespace,
        flatten=False)


class UssdRequest(object):
    """
    :param session_id:
        used to get session or create session if does not
        exits.

        If session is less than 8 we add *s* to make the session
        equal to 8

    :param phone_number:
        This the user identifier

    :param input:
        This ussd input the user has entered.

    :param language:
        Language to use to display ussd

    :param kwargs:
        Extra arguments.
        All the extra arguments will be set to the self attribute

        For instance:

        .. code-block:: python

            from ussd.core import UssdRequest

            ussdRequest = UssdRequest(
                '12345678', '702729654', '1', 'en',
                name='mwas'
            )

            # accessing kwarg argument
            ussdRequest.name
    """
    def __init__(self, session_id, phone_number,
                 ussd_input, language, default_language=None,
                 use_built_in_session_management=False,
                 expiry=180,
                 **kwargs):
        """
        :param session_id: Used to maintain session 
        :param phone_number: user dialing in   
        :param ussd_input: input entered by user
        :param language: language to be used
        :param default_language: language to used
        :param use_built_in_session_management: Used to enable ussd_airflow to 
            manage its own session, by default its set to False, is set to true 
        then the session_id should be None and expiry can't be None. 
        :param expiry: Its only used if use_built_in_session_management has
        been enabled. 
        :param kwargs: All other extra arguments
        """

        self.expiry = expiry
        # A bit of defensive programming to make sure
        # session_built_in_management has been initiated
        if use_built_in_session_management and session_id is not None:
            raise InvalidAttribute("When using built_in_session_management "
                                   "has been enabled session_id should "
                                   "be None")
        if use_built_in_session_management and expiry is None:
            raise InvalidAttribute("When built_in_session_management has been"
                                   "enabled expiry should not be None")
        # session id should not be None if built in session management
        # has not been enabled
        if session_id is None and not use_built_in_session_management:
            raise InvalidAttribute(
                "Session id should not be None if built in session management "
                "has not been enabled"
            )

        if use_built_in_session_management:
            session_id = self.get_or_create_session_id(phone_number)
        else:
            # if session id is less than 8 should provide the
            # suplimentary characters with 's'
            if len(str(
                    session_id)) < 8 and not use_built_in_session_management:
                session_id = 's' * (8 - len(str(session_id))) + session_id

        self.phone_number = phone_number
        self.input = unquote(ussd_input)
        self.language = language
        self.default_language = default_language or 'en'
        self.session_id = session_id
        self.session = ussd_session(self.session_id)

        for key, value in kwargs.items():
            setattr(self, key, value)


    def forward(self, handler_name):
        """
        Forwards a copy of the current request to a new
        handler. Clears any input, as it is assumed this was meant for
        the previous handler. If you need to pass info between
        handlers, do it through the USSD session.
        """
        new_request = copy(self)
        new_request.input = ''
        return new_request, handler_name

    def all_variables(self):
        all_variables = copy(self.__dict__)

        # delete session if it exist
        all_variables.pop("session", None)

        return all_variables

    def get_or_create_session_id(self, user_id):
        session_mapping = get_object_or_None(SessionLookup, user_id=user_id)

        # if its missing create a new one.
        if session_mapping is None:
            session_mapping = SessionLookup.objects.create(
                user_id=user_id,
                session_id=generate_session_id()
            )
        else:
            session = ussd_session(session_mapping.session_id)

            # get last time session was updated
            if session.get(ussd_airflow_variables.last_update):
                last_updated = utilities.string_to_datetime(
                    session[ussd_airflow_variables.last_update])
            else:
                last_updated = timezone.make_naive(session_mapping.updated_at)

            # check inactivity or if session has been closed
            inactivity_duration = (datetime.now() - last_updated).total_seconds()
            if inactivity_duration > self.expiry or \
                    session.get(ussd_airflow_variables.expiry):

                # update session_mapping with the new session_id
                session_mapping.session_id = generate_session_id()
                session_mapping.save()

        return session_mapping.session_id




class UssdResponse(object):
    """
    :param text:
        This is the ussd text to display to the user
    :param status:
        This shows the status of ussd session.

        True -> to continue with the session

        False -> to end the session
    :param session:
        This is the session object of the ussd session
    """
    def __init__(self, text, status=True, session=None):
        self.text = text
        self.status = status
        self.session = session

    def dumps(self):
        return self.text

    def __str__(self):
        return self.dumps()


class UssdHandlerMetaClass(type):

    def __init__(cls, name, bases, attr, **kwargs):
        super(UssdHandlerMetaClass, cls).__init__(
            name, bases, attr)

        abstract = attr.get('abstract', False)

        if not abstract or attr.get('screen_type', '') == 'custom_screen':
            required_attributes = ('screen_type', 'serializer', 'handle')

            # check all attributes have been defined
            for attribute in required_attributes:
                if attribute not in attr and not hasattr(cls, attribute):
                    raise MissingAttribute(
                        "{0} is required in class {1}".format(
                            attribute, name)
                    )

            if not isinstance(attr['serializer'], SerializerMetaclass):
                raise InvalidAttribute(
                    "serializer should be a "
                    "instance of {serializer}".format(
                        serializer=SerializerMetaclass)
                )
            _registered_ussd_handlers[attr['screen_type']] = cls


class UssdHandlerAbstract(object, metaclass=UssdHandlerMetaClass):
    abstract = True

    def __init__(self, ussd_request: UssdRequest,
                 handler: str, screen_content: dict,
                 initial_screen: dict, logger=None):
        self.ussd_request = ussd_request
        self.handler = handler
        self.screen_content = screen_content

        self.SINGLE_VAR = re.compile(r"^%s\s*(\w*)\s*%s$" % (
            '{{', '}}'))
        self.clean_regex = re.compile(r'^{{\s*(\S*)\s*}}$')
        self.logger = logger or get_logger(__name__).bind(
            handler=self.handler,
            screen_type=getattr(self, 'screen_type', 'custom_screen'),
            **ussd_request.all_variables(),
        )
        self.initial_screen = initial_screen

        self.pagination_config = self.initial_screen.get('pagination_config',
                                                         {})

        self.pagination_more_option = self._add_end_line(
            self.get_text(
                self.pagination_config.get('more_option', "more\n")
            )
        )
        self.pagination_back_option = self._add_end_line(
            self.get_text(
                self.pagination_config.get('back_option', "back\n")
            )
        )
        self.ussd_text_limit = self.pagination_config.\
            get("ussd_text_limit", ussd_airflow_variables.ussd_text_limit)

    def handle(self):
        if not self.ussd_request.input:
            ussd_response = self.show_ussd_content()
            return ussd_response if isinstance(ussd_response, UssdResponse) \
                else UssdResponse(str(ussd_response))
        return self.handle_ussd_input(self.ussd_request.input)

    def get_text_limit(self):
        return self.ussd_text_limit

    def show_ussd_content(self, **kwargs):
        raise NotImplementedError

    def handle_ussd_input(self, ussd_input):
        raise NotImplementedError

    @staticmethod
    def get_session_items(session) -> dict:
        return dict(iter(session.items()))

    @classmethod
    def get_context(cls, session, extra_context=None):
        context = cls.get_session_items(session)

        context.update(
            dict(os.environ)
        )

        if extra_context is not None:
            context.update(extra_context)

        # add timestamp in the context
        context.update(
            dict(now=datetime.now())
        )
        return context

    @staticmethod
    def render_text(session, text, context=None, extra=None, encode=None):
        if context is None:
            context = UssdHandlerAbstract.get_context(
                session
            )

        if extra:
            context.update(extra)

        template = env.from_string(text or '')
        text = template.render(context)
        return json.dumps(text) if encode is 'json' else text

    def get_text(self, text_context=None):
        text_context = self.screen_content.get('text')\
                       if text_context is None \
                       else text_context

        if isinstance(text_context, dict):
            language = self.ussd_request.language \
                   if self.ussd_request.language \
                          in text_context.keys() \
                   else self.ussd_request.default_language

            text_context = text_context[language]

        return self.render_text(
            self.ussd_request.session,
            text_context
        )

    @classmethod
    def evaluate_jija_expression(cls, expression, session,
                                 extra_context=None,
                                 lazy_evaluating=False):
        if not isinstance(expression, str) or \
                (lazy_evaluating and not cls._contains_vars(
                    expression)):
            return expression

        context = cls.get_context(
            session, extra_context=extra_context)

        expression = expression.replace("{{", "").replace("}}", "")
        try:
            expr = env.compile_expression(
                expression
            )
        except TemplateSyntaxError:
            return []
        return expr(context)

    @classmethod
    def validate(cls, screen_name: str, ussd_content: dict) -> (bool, dict):
        screen_content = ussd_content[screen_name]
        # adding screen name in context might be needed by validator
        ussd_content['screen_name'] = screen_name
        validation = cls.serializer(data=screen_content,
                                     context=ussd_content)
        del ussd_content['screen_name']
        if validation.is_valid():
            return True, {}
        return False, validation.errors

    @staticmethod
    def _contains_vars(data):
        '''
        returns True if the data contains a variable pattern
        '''
        if isinstance(data, str):
            for marker in ('{%', '{{', '{#'):
                if marker in data:
                    return True
        return False

    @staticmethod
    def _add_end_line(text):
        if text and '\n' not in text:
            text += '\n'
        return text

    def get_loop_items(self):
        loop_items = self.evaluate_jija_expression(
            self.screen_content["with_items"],
            session=self.ussd_request.session
        ) if self.screen_content.get("with_items") else [0] or [0]
        return loop_items

    @classmethod
    def render_request_conf(cls, session, data):
        if isinstance(data, str):
            return cls.render_text(session, data)

        elif isinstance(data, list):
            list_data = []
            for i in data:
                list_data.append(cls.render_request_conf(
                    session, i))

            return list_data

        elif isinstance(data, dict):
            dict_data = {}
            for key, value in data.items():
                dict_data.update(
                    {key: cls.render_request_conf(
                        session, value)}
                )
            return dict_data
        else:
            return data

    @staticmethod
    def get_variables_from_response_obj(response):
        response_varialbes = {}

        for i in inspect.getmembers(response):
            # Ignores anything starting with underscore
            # (that is, private and protected attributes)
            if not i[0].startswith('_'):
                # Ignores methods
                if not inspect.ismethod(i[1]) and \
                                type(i[1]) in \
                                (str, dict, int, dict, float, list, tuple):
                    if len(i) == 2:
                        response_varialbes.update(
                            {i[0]: i[1]}
                        )

        try:
            response_content = json.loads(response.content.decode())
        except json.JSONDecodeError:
            response_content = response.content.decode()

        if isinstance(response_content, dict):
            response_varialbes.update(
                response_content
            )

        # update content to save the one that has been decoded
        response_varialbes.update(
            {"content": response_content}
        )

        return response_varialbes
    @classmethod
    def make_request(cls, http_request_conf, response_session_key_save,
                     session, logger=None
                     ):
        logger = logger or get_logger(__name__).bind(
            action="make_request",
            session_id=session.session_key
        )
        logger.info("sending_request", **http_request_conf)
        response = requests.request(**http_request_conf)
        logger.info("response", status_code=response.status_code,
                         content=response.content)

        response_to_save = cls.get_variables_from_response_obj(response)

        # save response in session
        session[response_session_key_save] = response_to_save

        return response

    @staticmethod
    def fire_ussd_report_session_task(initial_screen: dict, session_id: str,
                            support_countdown=True):
        ussd_report_session = initial_screen['ussd_report_session']
        args = (session_id,)
        kwargs = {'screen_content': initial_screen}
        keyword_args = ussd_report_session.get("async_parameters",
                                               {"countdown": 900}).copy()
        if not support_countdown and keyword_args.get('countdown'):
            del keyword_args['countdown']

        report_session.apply_async(
            args=args,
            kwargs=kwargs,
            **keyword_args
        )


class UssdView(APIView):
    """
    To create Ussd View requires the following things:
        - Inherit from **UssdView** (Mandatory)
            .. code-block:: python

                from ussd.core import UssdView

        - Define Http method either **get** or **post** (Mandatory)
            The http method should return Ussd Request

                .. autoclass:: ussd.core.UssdRequest

        - define this varialbe *customer_journey_conf*
            This is the path of the file that has ussd screens
            If you want your file to be dynamic implement the
            following method **get_customer_journey_conf** it
            will be called by request object

        - define this variable *customer_journey_namespace*
            Ussd_airflow uses this namespace to save the
            customer journey content in memory. If you want
            customer_journey_namespace to be dynamic implement
            this method **get_customer_journey_namespace** it
            will be called with request object

        - override HttpResponse
            In ussd airflow the http method return UssdRequest object
            not Http response. Then ussd view gets UssdResponse object
            and convert it to HttpResponse. The default HttpResponse
            returned is a normal HttpResponse with body being ussd text

            To override HttpResponse returned define this method.
            **ussd_response_handler** it will be called with
            **UssdResponse** object.

                .. autoclass:: ussd.core.UssdResponse

    Example of Ussd view

    .. code-block:: python

        from ussd.core import UssdView, UssdRequest


        class SampleOne(UssdView):

            def get(self, req):
                return UssdRequest(
                    phone_number=req.data['phoneNumber'].strip('+'),
                    session_id=req.data['sessionId'],
                    ussd_input=text,
                    service_code=req.data['serviceCode'],
                    language=req.data.get('language', 'en')
                )

    Example of Ussd View that defines its own HttpResponse.

    .. code-block:: python

        from ussd.core import UssdView, UssdRequest


        class SampleOne(UssdView):

            def get(self, req):
                return UssdRequest(
                    phone_number=req.data['phoneNumber'].strip('+'),
                    session_id=req.data['sessionId'],
                    ussd_input=text,
                    service_code=req.data['serviceCode'],
                    language=req.data.get('language', 'en')
                )

            def ussd_response_handler(self, ussd_response):
                    if ussd_response.status:
                        res = 'CON' + ' ' + str(ussd_response)
                        response = HttpResponse(res)
                    else:
                        res = 'END' + ' ' + str(ussd_response)
                        response = HttpResponse(res)
                    return response
    """
    customer_journey_conf = None
    customer_journey_namespace = None

    def initial(self, request, *args, **kwargs):
        # initialize restframework
        super(UssdView, self).initial(request, args, kwargs)

        # initialize ussd
        self.ussd_initial(request)

    def ussd_initial(self, request, *args, **kwargs):
        if hasattr(self, 'get_customer_journey_conf'):
            self.customer_journey_conf = self.get_customer_journey_conf(
                request
            )
        if hasattr(self, 'get_customer_journey_namespace'):
            self.customer_journey_namespace = \
                self.get_customer_journey_namespace(request)

        if self.customer_journey_conf is None \
                or self.customer_journey_namespace is None:
            raise MissingAttribute("attribute customer_journey_conf and "
                                   "customer_journey_namespace are required")

        if not self.customer_journey_namespace in \
                staticconf.config.configuration_namespaces:
            load_yaml(
                self.customer_journey_conf,
                self.customer_journey_namespace
            )

        # confirm variable template has been loaded
        # get initial screen

        initial_screen = staticconf.read(
            "initial_screen",
            namespace=self.customer_journey_namespace)

        if isinstance(initial_screen, dict) and \
                initial_screen.get('variables'):
            variable_conf = initial_screen['variables']
            file_path = variable_conf['file']
            namespace = variable_conf['namespace']
            if not namespace in \
                    staticconf.config.configuration_namespaces:
                load_yaml(file_path, namespace)

        self.initial_screen = initial_screen \
            if isinstance(initial_screen, dict) \
            else {"initial_screen": initial_screen}

    def finalize_response(self, request, response, *args, **kwargs):

        if isinstance(response, UssdRequest):
            self.logger = get_logger(__name__).bind(**response.all_variables())
            try:
                ussd_response = self.ussd_dispatcher(response)
            except Exception as e:
                if settings.DEBUG:
                    ussd_response = UssdResponse(str(e))
            return self.ussd_response_handler(ussd_response)
        return super(UssdView, self).finalize_response(
            request, response, args, kwargs)

    def ussd_response_handler(self, ussd_response):
        return HttpResponse(str(ussd_response))

    def ussd_dispatcher(self, ussd_request):

        # Clear input and initialize session if we are starting up
        if '_ussd_state' not in ussd_request.session:
            ussd_request.input = ''
            ussd_request.session['_ussd_state'] = {'next_screen': ''}
            ussd_request.session['ussd_interaction'] = []
            ussd_request.session['posted'] = False
            ussd_request.session['submit_data'] = {}
            ussd_request.session['session_id'] = ussd_request.session_id
            ussd_request.session['phone_number'] = ussd_request.phone_number

        # update ussd_request variable to session and template variables
        # to be used later for jinja2 evaluation
        ussd_request.session.update(ussd_request.all_variables())

        # for backward compatibility
        # there are some jinja template using ussd_request
        # eg. {{ussd_request.session_id}}
        ussd_request.session.update(
            {"ussd_request": ussd_request.all_variables()}
        )

        self.logger.debug('gateway_request', text=ussd_request.input)

        # Invoke handlers
        ussd_response = self.run_handlers(ussd_request)
        ussd_request.session[ussd_airflow_variables.last_update] = \
            utilities.datetime_to_string(datetime.now())
        # Save session
        ussd_request.session.save()
        self.logger.debug('gateway_response', text=ussd_response.dumps(),
                     input="{redacted}")

        return ussd_response

    def run_handlers(self, ussd_request):

        handler = ussd_request.session['_ussd_state']['next_screen'] \
            if ussd_request.session.get('_ussd_state', {}).get('next_screen') \
            else "initial_screen"


        ussd_response = (ussd_request, handler)

        if handler != "initial_screen":
            # get start time
            start_time = utilities.string_to_datetime(
                ussd_request.session["ussd_interaction"][-1]["start_time"])
            end_time = datetime.now()
            # Report in milliseconds
            duration = (end_time - start_time).total_seconds() * 1000
            ussd_request.session["ussd_interaction"][-1].update(
                {
                    "input": ussd_request.input,
                    "end_time": utilities.datetime_to_string(end_time),
                    "duration": duration
                }
            )

        # Handle any forwarded Requests; loop until a Response is
        # eventually returned.
        while not isinstance(ussd_response, UssdResponse):
            ussd_request, handler = ussd_response

            screen_content = staticconf.read(
                handler,
                namespace=self.customer_journey_namespace)

            screen_type = 'initial_screen' \
                if handler == "initial_screen" and \
                   isinstance(screen_content, str) \
                else screen_content['type']

            ussd_response = _registered_ussd_handlers[screen_type](
                ussd_request,
                handler,
                screen_content,
                initial_screen=self.initial_screen,
                logger=self.logger
            ).handle()

        ussd_request.session['_ussd_state']['next_screen'] = handler

        ussd_request.session['ussd_interaction'].append(
            {
                "screen_name": handler,
                "screen_text": str(ussd_response),
                "input": ussd_request.input,
                "start_time": utilities.datetime_to_string(datetime.now())
            }
        )
        # Attach session to outgoing response
        ussd_response.session = ussd_request.session

        return ussd_response

    @staticmethod
    def validate_ussd_journey(ussd_content: dict) -> (bool, dict):
        errors = {}
        is_valid = True

        # should define initial screen
        if not 'initial_screen' in ussd_content.keys():
            is_valid = False
            errors.update(
                {'hidden_fields': {
                    "initial_screen": ["This field is required."]
                }}
            )
        for screen_name, screen_content in ussd_content.items():
            # all screens should have type attribute
            if screen_name == "initial_screen" and \
                    isinstance(screen_content, str):
                if not screen_content in ussd_content.keys():
                    is_valid = False
                    errors.update(
                        dict(
                            screen_name="Screen not available"
                        )
                    )
                continue

            screen_type = screen_content.get('type')

            # all screen should have type field.
            serialize = UssdBaseSerializer(data=screen_content,
                                           context=ussd_content)
            base_validation = serialize.is_valid()

            if serialize.errors:
                errors.update(
                    {screen_name: serialize.errors}
                )

            if not base_validation:
                is_valid = False
                continue

            # all screen type have their handlers
            handlers = _registered_ussd_handlers[screen_type]

            screen_validation, screen_errors = handlers.validate(
                screen_name,
                ussd_content
            )
            if screen_errors:
                errors.update(
                    {screen_name: screen_errors}
                )

            if not screen_validation:
                is_valid = screen_validation

        return is_valid, errors


import staticconf
from django.core.management.base import BaseCommand, CommandError
from ussd.core import UssdView
import os
import json


class Command(BaseCommand):
    help = 'Validate ussd customer journey'

    def add_arguments(self, parser):
        parser.add_argument('ussd_configs', nargs='+', type=str)

    def handle(self, *args, **options):
        error_message = {}
        for ussd_config in options["ussd_configs"]:
            if not os.path.isfile(ussd_config):
                raise CommandError("This file path {} does not exist".format(ussd_config))
            staticconf.YamlConfiguration(
                ussd_config,
                namespace="validation",
                flatten=False)

            ussd_screens = staticconf.config. \
                get_namespace("validation"). \
                get_config_values()

            is_valid, error_ussd_config_message = UssdView.validate_ussd_journey(
                ussd_screens)
            error_message[ussd_config] = dict(
                valid=is_valid,
                error_message=error_ussd_config_message
            )
            if not is_valid:
                raise CommandError(json.dumps(error_message))

        self.stdout.write(json.dumps(error_message))



from ussd.tests import UssdTestCase
from ussd.core import ussd_session


class TestUpdateSessionScreen(UssdTestCase.BaseUssdTestCase):
    validation_error_message = dict(
        screen_one=dict(
            next_screen=['This field is required.'],
            values_to_update=['This field is required.']
        ),
        screen_two=dict(
            values_to_update=dict(
                key=['This field is required.'],
                value=['This field is required.']
            )
        ),
        screen_three=dict(
            values_to_update=dict(
                key=['This field is required.'],
                value=['This field is required.']
            )
        )
    )

    def test(self):
        ussd_client = self.ussd_client()
        ussd_client.send('')
        session = ussd_session(ussd_client.session_id)

        self.assertEqual(
            session['customer_status'],
            "registered"
        )

        self.assertEqual(
            session['aged_24'],
            [
                {
                    'name': "Francis Mwangi",
                    'age': 24,
                    'height': 5.4
                },
                {
                    "name": "Wambui",
                    "age": 24,
                    "height": 5.4
                }
            ]
        )

        self.assertEqual(
            session["height_54"],
            [
                {
                    'name': "Francis Mwangi",
                    'age': 24,
                    'height': 5.4
                },
                {
                    'name': 'Isaac Karanja',
                    'age': 22,
                    'height': 5.4
                },
                {
                    "name": "Wambui",
                    "age": 24,
                    "height": 5.4
                }
            ]
        )
        # test for unregistered user
        ussd_client = self.ussd_client(phone_number=404)
        ussd_client.send('')
        session = ussd_session(ussd_client.session_id)

        self.assertEqual(
            session['customer_status'],
            "not_registered"
        )



from django.db import models


class SessionLookup(models.Model):
    """
    This model is used by built in session management 
    to map between user_id and session_id
    """
    user_id = models.CharField(max_length=255)  # this can also be phone number
    session_id = models.CharField(max_length=255)
    created_at = models.DateTimeField(u'Creation Date', auto_now=True)
    updated_at = models.DateTimeField(u'Update Date', auto_now_add=True)



from rest_framework import serializers


class UssdBaseSerializer(serializers.Serializer):
    """
    Each screen as a field "type" to determine which screen
    to display.

    type should be the supported only.
    for those that are not supported will get this message
       "Invalid screen type not supported"
    """
    type = serializers.CharField(max_length=50)

    def validate_type(self, value):
        # to avoid cyclic import
        from ussd.core import _registered_ussd_handlers
        if value not in _registered_ussd_handlers.keys():
            raise serializers.ValidationError("Invalid screen "
                                              "type not supported")
        return value


class UssdTextField(serializers.DictField):
    """
    Ussd text that's going to be displayed to the user.
    example

    text: This is the text that is going to be shown

    To display multi language

    text:
        en: This the text in english
        sw: this the text in swahili
        default: en
    """

    def to_internal_value(self, data):
        if not isinstance(data, dict):
            data = {
                'en': data,
                'default': 'en'}
        return super(UssdTextField, self).to_internal_value(data)

class UssdTextSerializer(serializers.Serializer):

    text = UssdTextField(child=serializers.CharField(max_length=500,
                                                     allow_blank=True))


class UssdContentBaseSerializer(UssdBaseSerializer, UssdTextSerializer):
    pass


class NextUssdScreenSerializer(serializers.Serializer):
    next_screen = serializers.CharField(max_length=50)

    def validate_next_screen(self, value):
        if value not in self.context.keys():
            raise serializers.ValidationError(
                "{screen} is missing in ussd journey".format(screen=value)
            )
        return value


class MenuOptionSerializer(UssdTextSerializer, NextUssdScreenSerializer):
    input_value = serializers.CharField(required=False, max_length=5,
                                        allow_blank=True)
    input_display = serializers.CharField(required=False, max_length=5)


class MenuSerializer(serializers.Serializer):
    options = serializers.ListField(
        child=MenuOptionSerializer(),
        required=True
    )


# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2017-03-08 15:06
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SessionLookup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('user_id', models.CharField(max_length=255)),
                ('session_id', models.CharField(max_length=255)),
                ('created_at', models.DateTimeField(auto_now=True, verbose_name='Creation Date')),
                ('updated_at', models.DateTimeField(auto_now_add=True, verbose_name='Update Date')),
            ],
        ),
    ]



from django.core.management import BaseCommand
from django.core.management import CommandError
from django.core.management import call_command
from ussd_airflow.settings import BASE_DIR

path = BASE_DIR + '/ussd_airflow_app_tpl'


class Command(BaseCommand):
    help = 'create ussd app'

    def add_arguments(self, parser):
        parser.add_argument('ussd_app_name', nargs='+', type=str)

    def handle(self, *args, **options):
        try:
            if len(options.get('ussd_app_name')[0].split()) > 1:
                raise CommandError
            app_name = options.get('ussd_app_name')[0]
            call_command('startapp', app_name, template=path)
        except CommandError:
            print('Provide a valid django App Name as documented here: '
                  ' https://docs.djangoproject.com/en/1.10/ref/django-admin/')



from setuptools import setup, find_packages
import os

__version__ = '0.0.4.1'


def _strip_comments(l):
    return l.split('#', 1)[0].strip()


def _pip_requirement(req):
    if req.startswith('-r '):
        _, path = req.split()
        return reqs(*path.split('/'))
    return [req]


def _reqs(*f):
    return [
        _pip_requirement(r) for r in (
            _strip_comments(l) for l in open(
                os.path.join(os.getcwd(), *f)).readlines()
        ) if r]


def reqs(*f):
    """Parse requirement file.
    Example:
        reqs('default.txt')          # requirements/default.txt
        reqs('extras', 'redis.txt')  # requirements/extras/redis.txt
    Returns:
        List[str]: list of requirements specified in the file.
    """
    return [req for subreq in _reqs(*f) for req in subreq]

setup(
    name='ussd_airflow',
    version=__version__,
    packages=find_packages(exclude=('ussd_airflow',)),
    url='https://github.com/mwaaas/ussd_airflow',
    install_requires=reqs('default.txt'),
    include_package_data=True,
    license='MIT',
    author='Mwas',
    author_email='mwasgize@gmail.com',
    description='Ussd Airflow Library'
)



from django.core.management import call_command
from django.test import LiveServerTestCase
import json
import os
from django.conf.urls import url, include
from django.test import TestCase

from ussd_airflow import urls
from rest_framework.test import APIClient


class TestUssdAppIsNotCreated(TestCase):

    def test_app_not_created(self):
        app_name = 'TestUssdApp TestUssdApp'
        call_command('create_ussd_app', app_name)
        with self.assertRaises(ImportError):
            include(app_name)


class TestUssdAppCreation(LiveServerTestCase):
    def setUp(self):
        self.client = APIClient()

    def test_app_creation(self):
        app_name = 'TestUssdApp'
        call_command('create_ussd_app', app_name)

        ussd_url = [
            url(r'^ussd/', include('TestUssdApp.urls'))
        ]

        urls.urlpatterns += ussd_url

        end_point_url = "http://localhost:8081/ussd/TestUssdApp_ussd_gateway"
        payload = {
            "phoneNumber": "0717199135", "sessionId": "12", "text": "1", "language": "en", "serviceCode": "200"}

        response = self.client.post(end_point_url,
                                    data=json.dumps(payload),
                                    content_type="application/json",)
        self.assertEqual(response.content,
                         b'END Example Quit Screen. Delete this and define your own customer journey screens.')
        self.assertEqual(response.status_code, 200)
        print(response.status_code)
        os.system('rm -r TestUssdApp')  # Remove created app. Clean up




import importlib
from datetime import datetime

date_format = "%Y-%m-%d %H:%M:%S.%f"


def str_to_class(import_path):
    module_name, class_name = import_path.rsplit(".", 1)
    try:
        module_ = importlib.import_module(module_name)
        try:
            class_ = getattr(module_, class_name)
        except AttributeError:
            raise Exception('Class does not exist')
    except ImportError:
        raise Exception('Module does not exist')
    return class_


def datetime_to_string(date_obj: datetime):
    return date_obj.strftime(date_format)


def string_to_datetime(date_str_obj: str):
    return datetime.strptime(date_str_obj, date_format)



from celery import current_app as app
import requests
from structlog import get_logger
from celery.exceptions import MaxRetriesExceededError


@app.task(bind=True)
def http_task(self, request_conf):
    requests.request(**request_conf)


@app.task(bind=True)
def report_session(self, session_id, screen_content):
    # to avoid circular import
    from ussd.core import ussd_session, UssdHandlerAbstract

    logger = get_logger(__name__).bind(
        action="report_session_task", session_id=session_id
    )

    logger.info('start')

    ussd_report_session_data = screen_content['ussd_report_session']

    session = ussd_session(session_id)

    if session.get('posted'):
        logger.info("session_already_reported", posted=session['posted'])
        return

    request_conf = UssdHandlerAbstract.render_request_conf(
        session,
        ussd_report_session_data['request_conf']
    )


    UssdHandlerAbstract.make_request(
        http_request_conf=request_conf,
        response_session_key_save=ussd_report_session_data['session_key'],
        session=session,
        logger=logger
    )

    # check if it is the desired effect
    for expr in ussd_report_session_data['validate_response']:
        if UssdHandlerAbstract.evaluate_jija_expression(
            expr['expression'],
            session=session
        ):
            session['posted'] = True
            session.save()
            return

    if ussd_report_session_data.get('retry_mechanism'):
        try:
            self.retry(**screen_content[
                    'ussd_report_session']['retry_mechanism'])
        except MaxRetriesExceededError as e:
            logger.warning("report_session_error", error_message=str(e))



from ussd.tests import UssdTestCase
from ussd.core import ussd_session


class TestRouterHandler(UssdTestCase.BaseUssdTestCase):
    validation_error_message = dict(
        invalid_router_1=dict(
            router_options=['This field is required.']
        ),
        invalid_router_2=dict(
            router_options=dict(
                next_screen=['This field is required.']
            )
        )
    )

    @staticmethod
    def add_phone_number_status_in_session(ussd_client):
        session = ussd_session(ussd_client.session_id)

        session["phone_numbers"] = {
            "203": ["registered"],
            "204": ["not_registered"],
            "205": ["not_there"]
        }
        session.save()

    def test(self):
        ussd_client = self.ussd_client(phone_number=200)

        self.assertEqual("This number is 200", ussd_client.send(''))

        ussd_client = self.ussd_client(phone_number=202)

        self.assertEqual("This number is 202", ussd_client.send(''))

        ussd_client = self.ussd_client(phone_number=300)

        self.assertEqual("This is the default screen",
                         ussd_client.send(''))

    def test_route_options_with_list_loop(self):
        ussd_client = self.ussd_client(phone_number=204)
        # add phone_number in session
        self.add_phone_number_status_in_session(ussd_client)

        # dial in
        response = ussd_client.send('')

        self.assertEqual(
            "You are not registered user",
            response
        )

        ussd_client = self.ussd_client(phone_number=203)
        # add phone_number in session
        self.add_phone_number_status_in_session(ussd_client)

        # dial in
        response = ussd_client.send('')

        self.assertEqual(
            "You are registered user",
            response
        )

        ussd_client = self.ussd_client(phone_number=205)
        # add phone_number in session
        self.add_phone_number_status_in_session(ussd_client)

        # dial in
        response = ussd_client.send('')

        self.assertEqual(
            "This is the default screen",
            response
        )

    def test_router_option_with_dict_loop(self):

        ussd_client = self.ussd_client(phone_number=206)

        self.assertEqual(
            "This is the default screen",
            ussd_client.send('')
        )

        ussd_client = self.ussd_client(phone_number=207)

        self.assertEqual(
            "This screen has been routed here because the "
            "phone number is 207",
            ussd_client.send('')
        )



"""
This tests creation of ussd screens using apis
"""

from django.test import LiveServerTestCase


class TestingFirstScenario(LiveServerTestCase):
    """
        test ussd by creating ussd screens by models instead of creating Handler classes

        # Test the following workflow

        A) welcome screen
        ---------------------------
         Welcome to renders:
                1. Request a loan
                2. Repay a loan
                3. About us
                4. Quit

        B) Option 1 selected
        --------------------------------

            Select one of the loans:
                1. Ksh 300
                2. Ksh 200
                3. Ksh 100
                4. Back

        C) Option 1 selected
        ----------------------------------

            Choose repayment plan
                1. KSH 350 in 7 days
                2. KSH 250 in 5 days
                3. Back

        D) Option 2 selected

            Confirm your loan application of Ksh 300 to be paid by
            interest of Ksh 330 in 7 days
                1. Confirm
                2. Back

        E) select option 1 ( to confirm the loan application)

            Thank you for your loan application . your loan will be
            processed and you will be notified



        The above workflow tests the following screens and functionality:
            - Menu screen
            - Menu and screen
            - Input screen
            - Quit screen

        example of menu lookup:
            {
            status=200,
            loan_offers=[300, 200, 100],
            repayment={300:{7:350, 5:250}, 200:{7: 300}, 100:{8:500} }
            }

        """
    pass